import wget
import hashlib
import os
import time
import requests
from geolite2 import geolite2
import sys
import subprocess
from .GlassConfig import GlassConfig

def md5(file_name):
    if not os.path.isfile(file_name):
        return ""

    md5_hash = hashlib.md5()
    content = open(file_name, "rb").read()
    md5_hash.update(content)

    return md5_hash.hexdigest()

def download(url, target_file, md5_str:str=""):
    target_folder = os.path.dirname(os.path.abspath(target_file))
    if not os.path.isdir(target_folder):
        os.makedirs(target_folder)

    if md5_str:
        current_md5 = md5(target_file)
        if current_md5 == md5_str:
            return

    times = 0
    while True:
        if os.path.isfile(target_file):
            os.remove(target_file)

        print(f"downloading {os.path.basename(target_file)}...")
        try:
            wget.download(url, target_file)
        except:
            respond = requests.get(url)
            if respond.status_code != 200:
                raise RuntimeError(respond.content)
            
            out_file = open(target_file, "wb")
            out_file.write(respond.content)
            out_file.close()

        if md5_str:
            current_md5 = md5(target_file)
            if current_md5 == md5_str:
                return
        else:
            if os.path.isfile(target_file):
                return

        time.sleep(1)

        times += 1
        if times > 4:
            raise RuntimeError(f"download {url} to {target_file} retry over times")
        else:
            print("download failed, retry...")

def public_ip():
    try:
        response = requests.get('https://httpbin.org/ip')
        return response.json().get('origin')
    except:
        return "127.0.0.1"

def is_China_ip(ip_address):
    if ip_address == "127.0.0.1":
        return True

    reader = geolite2.reader()
    location = reader.get(ip_address)
    country = location.get('country', {}).get('iso_code') if location else None
    return country == 'CN'

def is_China_user():
    return is_China_ip(public_ip())

def pip_install(package_name:str):
    if "/" not in package_name:
        install_cmd = [sys.executable, "-m", "pip", "install", package_name]
        if is_China_user():
            install_cmd = [sys.executable, "-m", "pip", "install", package_name, "-i", "https://pypi.tuna.tsinghua.edu.cn/simple"]
        
        return_code = subprocess.call(install_cmd)
        if return_code != 0:
            raise RuntimeError(f"failed to install {package_name}")
    else:
        target_file = GlassConfig.cache_folder + "/" + os.path.basename(package_name)
        download(package_name, target_file)
        install_cmd = [sys.executable, "-m", "pip", "install", target_file]
        return_code = subprocess.call(install_cmd)
        if return_code != 0:
            raise RuntimeError(f"failed to install {package_name}")


import copy
import numpy as np
import glm

from .utils import capacity_of
from .helper import nitems
from .GlassConfig import GlassConfig
from .GLInfo import GLInfo

class SameTypeList:

    class iterator:
        def __init__(self, _list):
            self.__list = _list
            self.__len_list = len(_list)
            self.__current_index = 0

        def __next__(self):
            if self.__current_index >= self.__len_list:
                raise StopIteration()
            
            value = self.__list[self.__current_index]
            self.__current_index += 1

            return value
        
        def __iter__(self):
            self.__current_index = 0
            return self
        
    class const_iterator:
        def __init__(self, _list):
            self.__list = _list
            self.__len_list = len(_list)
            self.__current_index = 0

        def __next__(self):
            if self.__current_index >= self.__len_list:
                raise StopIteration()
            
            value = self.__list.const_get(self.__current_index)
            self.__current_index += 1

            return value
        
        def __iter__(self):
            self.__current_index = 0
            return self

    def __init__(self, _list:(list,np.ndarray)=None, dtype:type=None):
        self.reset(_list, dtype)

    def reset(self, _list:(list,np.ndarray)=None, dtype:type=None):
        if _list is None:
            _list = []

        self._list = _list
        self._list_ndarray = None
        self._list_dirty = True
        self._should_retest = True
        self._data_list = []

        str_dtype = str(dtype)
        if str_dtype == "callback_vec3":
            dtype = glm.vec3
        elif str_dtype == "callback_vec4":
            dtype = glm.vec4
        elif str_dtype == "callback_quat":
            dtype = glm.quat

        self._dtype = dtype
        if dtype is None and _list:
            self._dtype = type(_list[0])

        self._increment = None
        self._checked_out_items = {}

    @property
    def ndarray(self)->np.ndarray:
        if self._list_dirty:
            if isinstance(self._list, np.ndarray):
                self._list_ndarray = self._list
            else:
                if self and "sampler" in type(self._list[0]).__name__:
                    len_list = len(self._list)
                    self._data_list.extend([0]*(len_list - len(self._data_list)))
                    del self._data_list[len_list:]
                    for i in range(len_list):
                        self._data_list[i] = self._list[i].handle

                    self._list_ndarray = np.array(self._data_list, dtype=np.uint64)
                else:
                    if self._dtype is None:
                        self._list_ndarray = np.array(self._list)
                    else:
                        np_dtype = self._dtype
                        if np_dtype in GLInfo.np_dtype_map:
                            np_dtype = GLInfo.np_dtype_map[np_dtype]
                        self._list_ndarray = np.array(self._list, dtype=np_dtype)

            self._list_dirty = False

        return self._list_ndarray

    @ndarray.setter
    def ndarray(self, array:np.ndarray):
        self._list_dirty = False
        self._should_retest = True
        self._list_ndarray = array
        self._list = array

        if self._increment is not None:
            _list = list(map(lambda x : self._dtype(*x), self._list))
            self._increment.reset(_list)

    @classmethod
    def frombuffer(cls, buffer:(bytes,bytearray), dtype):
        np_dtype = GLInfo.np_dtype_map[dtype]
        np_array = np.frombuffer(buffer, dtype=np_dtype).reshape((-1, nitems(dtype)))
        result = cls(np_array, dtype=dtype)
        return result

    @classmethod
    def fromarray(cls, np_array:np.ndarray, dtype):
        np_array = np_array.reshape((-1, nitems(dtype)))
        result = cls(np_array, dtype=dtype)
        return result

    def _check_type(self, value):
        if not self:
            return

        if type(value) != self._dtype and type(value) != type(self._list[0]):
            raise TypeError(f"{self._dtype} value expected, {type(value)} value was given")

    def _change_to_list(self):
        if not isinstance(self._list, list):
            self._list = list(map(lambda x : self._dtype(*x), self._list))

    def append(self, value):
        self._change_to_list()
        self._check_in_items()

        if GlassConfig.debug:
            self._check_type(value)

        self._list.append(value)

        if self._increment is not None:
            self._increment.append(value)

        self._list_dirty = True
        self._should_retest = True

    def extend(self, _list):
        self._change_to_list()
        self._check_in_items()

        if GlassConfig.debug:
            for value in _list:
                self._check_type(value)

        self._list.extend(_list)

        if self._increment is not None:
            self._increment.append(_list)

        self._list_dirty = True
        self._should_retest = True

    def insert(self, index, value):
        self._change_to_list()
        self._check_in_items()

        if GlassConfig.debug:
            self._check_type(value)

        self._list.insert(index, value)

        if self._increment is not None:
            self._increment.insert(index, value)

        self._list_dirty = True
        self._should_retest = True

    def remove(self, value):
        self._change_to_list()
        self._check_in_items()

        index = self.index(value)

        self._list.remove(value)

        if self._increment is not None:
            self._increment.delete(index)

        self._list_dirty = True
        self._should_retest = True

    def pop(self, index:int):
        self._change_to_list()
        self._check_in_items()

        value = self._list.pop(index)

        if self._increment is not None:
            self._increment.delete(index)

        self._list_dirty = True
        self._should_retest = True

        return value

    def clear(self):
        self._check_in_items()
        self._list = []

        if self._increment is not None:
            self._increment.clear()

        self._list_dirty = True
        self._should_retest = True

    def update(self, _list):
        len_list = len(_list)
        len_self = len(self._list)
        if len_list < len_self:
            del self[len_list:]
        
        for i in range(len_self):
            self[i] = _list[i]

        if len_list > len_self:
            self.extend(_list[len_self:])

        self._list_dirty = True
        self._should_retest = True

    def __len__(self):
        if isinstance(self._list, np.ndarray):
            return self._list.shape[0]
        else:
            return len(self._list)
    
    def __bool__(self):
        return (len(self._list) > 0)

    def __setitem__(self, index, value):
        self._check_in_items()

        if isinstance(index, int):
            len_list = len(self._list)
            if index >= len_list:
                self.extend([value]*(index+1-len_list))
                self._list_dirty = True
                self._should_retest = True
                return True
            
            if self.const_get(index) == value:
                return False

        if GlassConfig.debug:
            if isinstance(index, slice) and type(value) != type(self._list[0]) and type(value) != self._dtype:
                for subvalue in value:
                    self._check_type(subvalue)
            else:
                self._check_type(value)

        self._list[index] = value
        if self._increment is not None:
            self._increment.update(index, value)

        self._list_dirty = True
        self._should_retest = True
        return True

    def __getitem__(self, index:(int,slice)):
        if self._increment is not None:
            if isinstance(index, slice):
                start, stop, step = self.__process_slice(index)
                for sub_index in range(start, stop, step):
                    if sub_index not in self._checked_out_items:
                        self._checked_out_items[sub_index] = copy.deepcopy(self.const_get(sub_index))
            else:
                if index not in self._checked_out_items:
                    self._checked_out_items[index] = copy.deepcopy(self.const_get(index))

        self._list_dirty = True
        self._should_retest = True
        return self.const_get(index)
    
    def const_get(self, index):
        result = self._list[index]
        if not isinstance(self._list, list):
            result = self._dtype(*result)
        return result

    def const_items(self):
        return SameTypeList.const_iterator(self)

    def __iter__(self):
        return SameTypeList.iterator(self)

    def _check_in_items(self):
        if self._increment is None or not self._checked_out_items:
            return
        
        for index, value in self._checked_out_items.items():
            new_value = self.const_get(index)
            if new_value != value:
                self._increment.update(index, new_value)

        self._checked_out_items.clear()

    def __delitem__(self, index:(int,slice)):
        self._change_to_list()
        self._check_in_items()
        del self._list[index]

        if self._increment is not None:
            if isinstance(index, slice):
                start, stop, step = self.__process_slice(index)
                if start >= len(self._list) or step == 0 or start == stop or \
                   (step > 0 and stop < step) or \
                   (step < 0 and stop > step):
                    return

                if step == 1:
                    self._increment.delete(start, stop)
                else:
                    for i in range(start, stop, step):
                        self._increment.delete(i)
            else:
                self._increment.delete(index)

        self._list_dirty = True
        self._should_retest = True

    def __process_slice(self, index):
        len_self = len(self._list)
        start = index.start if index.start is not None else 0
        stop = index.stop if index.stop is not None else len_self
        step = index.step if index.step is not None else 1

        if len_self > 0:
            while start < 0:
                start += len_self

            while stop < 0:
                stop += len_self

        if start < 0: start = 0
        if stop > len_self: stop = len_self

        return start, stop, step

    @property
    def increment(self):
        return self._increment

    @property
    def capacity(self):
        return capacity_of(len(self._list))
    
    @property
    def dtype(self):
        if self._dtype is not None:
            return self._dtype
        
        if self:
            return type(self._list[0])
        
        return None

    @dtype.setter
    def dtype(self, dtype):
        str_dtype = str(dtype)
        if str_dtype == "callback_vec3":
            dtype = glm.vec3
        elif str_dtype == "callback_vec4":
            dtype = glm.vec4
        elif str_dtype == "callback_quat":
            dtype = glm.quat

        if self._dtype == dtype:
            return
        
        self._dtype = dtype
        self._list_dirty = True
        self._should_retest = True


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel, QComboBox, QColorDialog, QFileDialog
from PyQt6.QtGui import QColor, QMouseEvent, QPaintEvent, QPainter, QPixmap
from PyQt6.QtCore import pyqtSignal, Qt, QRect

from enum import Enum
import glm
import os

class ColorImageChooser(QWidget):

    class Type(Enum):
        Color = 0
        Image = 1

    type_changed = pyqtSignal(Type)
    color_changed = pyqtSignal(QColor)
    glm_color_changed = pyqtSignal(object)
    file_path_changed = pyqtSignal(str)

    def __init__(self, type:Type=Type.Color, width:int=35, height:int=35, alpha:bool=False, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)
        self.__type:ColorImageChooser.Type = type
        self.__alpha = alpha
        self.__color:QColor = QColor(0,0,0)
        self.__file_path:str = ""
        self.setFixedSize(width, height)
    
    @property
    def type(self):
        return self.__type
    
    @type.setter
    def type(self, type:Type):
        if self.__type == type:
            return
        
        self.__type = type
        self.update()
        self.type_changed.emit(self.__type)

    @property
    def color(self):
        return self.__color
    
    @color.setter
    def color(self, color:QColor):
        if self.__color == color:
            return
        
        self.__color = color
        if self.__type == ColorImageChooser.Type.Color:
            self.update()
        self.color_changed.emit(self.__color)
        self.glm_color_changed.emit(self.glm_color)

    @property
    def glm_color(self):
        if self.__alpha:
            return glm.vec4(self.__color.redF(), self.__color.greenF(), self.__color.blueF(), self.__color.alphaF())
        else:
            return glm.vec3(self.__color.redF(), self.__color.greenF(), self.__color.blueF())

    @glm_color.setter
    def glm_color(self, glm_color):
        if isinstance(glm_color, glm.vec3):
            self.color = QColor(int(glm_color.r*255), int(glm_color.g*255), int(glm_color.b*255))
        else:
            self.color = QColor(int(glm_color.r*255), int(glm_color.g*255), int(glm_color.b*255), int(glm_color.a*255))

    @property
    def file_path(self):
        return self.__file_path
    
    @file_path.setter
    def file_path(self, file_path:str):
        if self.__file_path == file_path:
            return
        
        self.__file_path = file_path
        if self.__type == ColorImageChooser.Type.Image:
            self.update()
        self.file_path_changed.emit(self.__file_path)

    def mouseReleaseEvent(self, mouse_event: QMouseEvent | None) -> None:
        if self.__type == ColorImageChooser.Type.Color:
            color = None
            if self.__alpha:
                color = QColorDialog.getColor(self.__color, self, "选择颜色", QColorDialog.ColorDialogOption.ShowAlphaChannel)
            else:
                color = QColorDialog.getColor(self.__color, self, "选择颜色")

            if color.isValid():
                self.color = color
        else:
            directory = None
            if os.path.isfile(self.__file_path):
                directory = os.path.dirname(self.__file_path)

            file_path = QFileDialog.getOpenFileName(self, "选择图片", directory, "Image File(*.jpg *.png *.bmp *.ico *.exr *.hdr *.glsl)")
            if file_path[0]:
                self.file_path = file_path[0]

        return super().mouseReleaseEvent(mouse_event)
    
    def paintEvent(self, paint_event: QPaintEvent | None) -> None:
        painter = QPainter(self)
        painter.setPen(Qt.PenStyle.NoPen)
        if self.__type == ColorImageChooser.Type.Color:
            painter.setBrush(self.__color)
        else:
            painter.setBrush(QColor(0,0,0))

        rect = self.rect()
        painter.drawRect(rect)
        if self.__type == ColorImageChooser.Type.Image and \
           os.path.isfile(self.__file_path):
            pixmap = QPixmap(self.__file_path)
            pixmap_ratio = pixmap.width() / pixmap.height()
            rect_ratio = rect.width() / rect.height()
            if pixmap_ratio > rect_ratio:
                pixmap_width = rect.width()
                pixmap_height = int(rect.width() / pixmap_ratio)
                contain_rect = QRect(0, int(rect.height()/2-pixmap_height/2), pixmap_width, pixmap_height)
                painter.drawPixmap(contain_rect, pixmap)
            else:
                pixmap_height = rect.height()
                pixmap_width = int(rect.height() * pixmap_ratio)
                contain_rect = QRect(int(rect.width()/2-pixmap_width/2), 0, pixmap_width, pixmap_height)
                painter.drawPixmap(contain_rect, pixmap)

        return super().paintEvent(paint_event)

class ColorImageInput(QWidget):

    type_changed = pyqtSignal(ColorImageChooser.Type)
    color_changed = pyqtSignal(QColor)
    glm_color_changed = pyqtSignal(glm.vec3)
    file_path_changed = pyqtSignal(str)

    def __init__(self, prompt:str, width:int=35, height:int=35, alpha:bool=False, parent:QWidget=None):
        QWidget.__init__(self, parent)

        hlayout = QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        label = QLabel(prompt, self)
        hlayout.addWidget(label)

        self.__type_combobox = QComboBox(self)
        self.__type_combobox.addItem("颜色", ColorImageChooser.Type.Color)
        self.__type_combobox.addItem("贴图", ColorImageChooser.Type.Image)
        hlayout.addWidget(self.__type_combobox)

        self.__color_image_chooser = ColorImageChooser(width=width, height=height, alpha=alpha, parent=self)
        hlayout.addWidget(self.__color_image_chooser)
        self.setLayout(hlayout)

        self.__color_image_chooser.type_changed.connect(self.type_changed)
        self.__color_image_chooser.color_changed.connect(self.color_changed)
        self.__color_image_chooser.glm_color_changed.connect(self.glm_color_changed)
        self.__color_image_chooser.file_path_changed.connect(self.file_path_changed)
        self.__type_combobox.currentIndexChanged.connect(self.__slot_type_changed)
        

    def __slot_type_changed(self, index:int):
        self.__color_image_chooser.type = self.__type_combobox.itemData(index)

    @property
    def type(self):
        return  self.__color_image_chooser.type
    
    @type.setter
    def type(self, type:ColorImageChooser.Type):
        if type == ColorImageChooser.Type.Color:
            self.__type_combobox.setCurrentIndex(0)
        else:
            self.__type_combobox.setCurrentIndex(1)

    @property
    def color(self):
        return self.__color_image_chooser.color
    
    @color.setter
    def color(self, color:QColor):
        self.__color_image_chooser.color = color

    @property
    def glm_color(self):
        return self.__color_image_chooser.glm_color
    
    @glm_color.setter
    def glm_color(self, glm_color:glm.vec3):
        self.__color_image_chooser.glm_color = glm_color

    @property
    def file_path(self):
        return self.__color_image_chooser.file_path
    
    @file_path.setter
    def file_path(self, file_path:str):
        self.__color_image_chooser.file_path = file_path


from ..Mesh import Mesh
from ..algorithm import line_intersect_plane, \
    point_mirror_by_plane, polygon_centroid, polygon_normal, cos_angle_of

from glass import Vertex

import glm
import math
from enum import Enum
import copy

class Extruder(Mesh):

    class JoinStyle(Enum):
        MiterJoin = 0
        RoundJoin = 1
        BevelJoin = 2

    def __init__(self, section, path, join_style:JoinStyle=JoinStyle.MiterJoin,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 n_corner_divide:int=100,
                 normalize_tex_coord:bool=False,
                 name:str="", block:bool=True, surf_type:Mesh.SurfType=Mesh.SurfType.Auto):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block, surf_type=surf_type)
        self.__section = section
        self.__path = path
        self.__join_style = join_style
        self.__n_corner_divide = n_corner_divide
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        section = self.__section
        path = copy.copy(self.__path)
        i = 1

        len_path = len(path)
        while i < len_path:
            while glm.length(path[i]-path[i-1]) < 1E-6:
                del path[i]
                len_path -= 1

            i += 1

        if len_path < 2:
            self.vertices.clear()
            self.indices.clear()
            return
        
        centroid = polygon_centroid(section)
        normal = -polygon_normal(section, centroid=centroid)
        section = self.__used_section

        dem = math.sqrt(normal.x**2 + normal.y**2)
        current_yaw = 0
        current_pitch = math.pi/2 if normal.z > 0 else -math.pi/2
        if dem > 1E-6:
            current_pitch = math.atan(normal.z/dem)
            current_yaw = -math.atan2(normal.x, normal.y)

        d2 = path[1] - path[0]
        dem = math.sqrt(d2.x**2 + d2.y**2)
        dest_yaw = 0
        dest_pitch = math.pi/2 if d2.z > 0 else -math.pi/2
        if dem > 1E-6:
            dest_pitch = math.atan(d2.z/dem)
            dest_yaw = -math.atan2(d2.x, d2.y)
        
        delta_yaw = dest_yaw - current_yaw
        delta_pitch = dest_pitch - current_pitch

        quat1 = glm.quat(math.cos(delta_yaw/2), 0, 0, math.sin(delta_yaw/2))
        quat2 = glm.quat(math.cos(delta_pitch/2), math.sin(delta_pitch/2), 0, 0)
        quat = quat1 * quat2

        section_offset = []
        current_section = []
        accumulate_length = [0]
        for j in range(len(section)):
            current_section.append(quat * (section[j]-centroid) + path[0])
            section_offset.append(glm.dot(section[j]-centroid, normal))
            if j > 0:
                accumulate_length.append(accumulate_length[-1] + glm.length(section[j] - section[j-1]))

        if self.__join_style == Extruder.JoinStyle.MiterJoin:
            yield from self.miter_join_build(current_section, accumulate_length, section_offset, path)
        elif self.__join_style == Extruder.JoinStyle.RoundJoin:
            yield from self.round_join_build(current_section, accumulate_length, section_offset, path)
        elif self.__join_style == Extruder.JoinStyle.BevelJoin:
            yield from self.bevel_join_build(current_section, accumulate_length, section_offset, path)

    def bevel_join_build(self, current_section, accumulate_length, section_offset, path):
        vertices = self.vertices
        indices = self.indices
        normalize_tex_coord = self.__normalize_tex_coord
        len_section = len(current_section)

        i_vertex = 0
        i_index = 0

        L = 0
        longest_distance = None
        next_section = []
        span_distances = []
        center_length = []
        d1, d2 = None, None
        len_path = len(path)
        for i in range(len_path):
            should_add_corner = True
            if i == 0:
                d1 = glm.normalize(path[i+1] - path[i])
                should_add_corner = False
            else:
                d1 = path[i] - path[i-1]
                len_d1 = glm.length(d1)
                d1 = d1 / len_d1

                L += len_d1
                if i < len_path-1:
                    d2 = glm.normalize(path[i+1] - path[i])
                    cos_theta = glm.dot(d1, d2)
                    if cos_theta > 0.9:
                        should_add_corner = False

                    half_way = glm.normalize(-d1 + d2)
                    plane_normal = None
                    if glm.length(-d1 + d2) < 1E-6:
                        plane_normal = glm.normalize(d1 + d2)
                    else:
                        up = glm.normalize(glm.cross(d1, d2))
                        plane_normal = glm.normalize(glm.cross(half_way, up))

                    section_on_bisection = []
                    for j in range(len_section):
                        point_on_bisection = line_intersect_plane(current_section[j], d1, path[i], plane_normal)
                        section_on_bisection.append(point_on_bisection)

                    if should_add_corner:
                        next_section = []
                        span_distances = []
                        center_length = [0]
                        last_center = None
                        longest_distance = -float("inf")
                        for j in range(len_section):
                            point_on_pendicular = line_intersect_plane(current_section[j], d1, path[i], d1)
                            point_on_bisection = section_on_bisection[j]

                            bisection_point_projection = glm.dot((point_on_bisection - path[i]), d1)
                            pendicular_point_projection = glm.dot((point_on_pendicular - path[i]), d1)

                            current_center = None
                            if bisection_point_projection < pendicular_point_projection:
                                current_section[j] = point_on_bisection
                                next_section.append(point_on_bisection)
                                span_distances.append(0)
                                current_center = point_on_bisection
                            else:
                                current_section[j] = point_on_pendicular
                                next_point = point_mirror_by_plane(point_on_pendicular, path[i], plane_normal)
                                next_section.append(next_point)
                                current_distance = glm.length(next_point - current_section[j])
                                span_distances.append(current_distance)
                                if current_distance > longest_distance:
                                    longest_distance = current_distance
                                current_center = 0.5*(next_point + current_section[j])
                            if j > 0:
                                center_length.append(center_length[-1] + glm.length(current_center - last_center))

                            last_center = current_center
                    else:
                        for j in range(len_section):
                            current_section[j] = section_on_bisection[j]
                else:
                    should_add_corner = False
                    for j in range(len_section):
                        current_section[j] = line_intersect_plane(current_section[j], d1, path[i], d1) + section_offset[j]*d1

            for j in range(len_section):
                t = accumulate_length[j]
                s = L + glm.dot(current_section[j] - path[i], d1)
                if normalize_tex_coord:
                    s /= accumulate_length[-1]
                    t /= accumulate_length[-1]

                vertex = Vertex()
                vertex.position = current_section[j]
                vertex.tex_coord = glm.vec3(s, t, 0)
                vertices[i_vertex] = vertex
                i_vertex += 1

                if j > 0 and i > 0:
                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 2 - len_section
                    triangle[2] = i_vertex - 2
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 1 - len_section
                    triangle[2] = i_vertex - 2 - len_section
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    yield

            if should_add_corner:
                for j in range(len_section):
                    s = L-span_distances[j]/2
                    t = center_length[j]
                    if normalize_tex_coord:
                        s /= center_length[-1]
                        t /= center_length[-1]

                    vertex = Vertex()
                    vertex.position = current_section[j]
                    vertex.tex_coord = glm.vec3(s, t, 0)
                    vertices[i_vertex] = vertex
                    i_vertex += 1

                for j in range(len_section):
                    s = L+span_distances[j]/2
                    t = center_length[j]
                    if normalize_tex_coord:
                        s /= center_length[-1]
                        t /= center_length[-1]

                    vertex = Vertex()
                    vertex.position = next_section[j]
                    vertex.tex_coord = glm.vec3(s, t, 0)
                    vertices[i_vertex] = vertex
                    i_vertex += 1

                    if j > 0 and i > 0:
                        triangle = glm.uvec3(0, 0, 0)
                        triangle[0] = i_vertex - 1
                        triangle[1] = i_vertex - 2 - len_section
                        triangle[2] = i_vertex - 2
                        indices[i_index] = triangle
                        i_index += 1
                        self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                        triangle = glm.uvec3(0, 0, 0)
                        triangle[0] = i_vertex - 1
                        triangle[1] = i_vertex - 1 - len_section
                        triangle[2] = i_vertex - 2 - len_section
                        indices[i_index] = triangle
                        i_index += 1
                        self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])
                    
                        yield

                for j in range(len_section):
                    s = L + glm.dot(next_section[j] - path[i], d2)
                    t = accumulate_length[j]
                    if normalize_tex_coord:
                        s /= accumulate_length[-1]
                        t /= accumulate_length[-1]

                    vertex = Vertex()
                    vertex.position = next_section[j]
                    vertex.tex_coord = glm.vec3(s, t, 0)
                    vertices[i_vertex] = vertex
                    i_vertex += 1

                current_section = section_on_bisection

        del vertices[i_vertex:]
        del indices[i_index:]
            
    def round_join_build(self, current_section, accumulate_length, section_offset, path):
        vertices = self.vertices
        indices = self.indices
        n_corner_divide = self.__n_corner_divide
        normalize_tex_coord = self.__normalize_tex_coord
        len_section = len(current_section)

        i_vertex = 0
        i_index = 0

        L = 0
        up = None
        first_hit_point = None
        distance_first_hit_to_corner = None
        theta = None
        backup_section = None
        last_should_add_corner = False
        len_path = len(path)
        for i in range(len_path):
            should_add_corner = True
            if i == 0:
                d1 = glm.normalize(path[i+1] - path[i])
                should_add_corner = False
            else:
                d1 = path[i] - path[i-1]
                len_d1 = glm.length(d1)
                d1 = d1 / len_d1
                
                if last_should_add_corner:
                    len_d1 -= distance_first_hit_to_corner*math.sin(theta/2)

                L += len_d1
                if i < len_path-1:
                    d2 = glm.normalize(path[i+1] - path[i])
                    cos_theta = glm.dot(d1, d2)
                    if cos_theta > 0.9:
                        should_add_corner = False
                    else:
                        theta = math.acos(cos_theta)

                    half_way = glm.normalize(-d1 + d2)
                    plane_normal = None
                    if glm.length(-d1 + d2) < 1E-6:
                        plane_normal = glm.normalize(d1 + d2)
                    else:
                        up = glm.normalize(glm.cross(d1, d2))
                        plane_normal = glm.normalize(glm.cross(half_way, up))

                    backup_section = []
                    for j in range(len_section):
                        backup_section.append(line_intersect_plane(current_section[j], d1, path[i], plane_normal))

                    if should_add_corner:
                        min_length = float("inf")
                        first_hit_point = None
                        for j in range(len_section):
                            current_point = backup_section[j]
                            current_length = glm.dot(current_point - path[i], d1)
                            if current_length < min_length:
                                min_length = current_length
                                first_hit_point = current_point
                        distance_first_hit_to_corner = glm.length(first_hit_point - path[i])
                        
                        for j in range(len_section):
                            current_section[j] = line_intersect_plane(current_section[j], d1, first_hit_point, d1)
                    
                    else:
                        for j in range(len_section):
                            current_section[j] = backup_section[j]
                else:
                    should_add_corner = False

                    for j in range(len_section):
                        current_section[j] = line_intersect_plane(current_section[j], d1, path[i], d1) + section_offset[j]*d1

            quat = None

            if should_add_corner:
                L -= distance_first_hit_to_corner*math.sin(theta/2)
                max_k = min(n_corner_divide, int(theta/0.05))
                for k in range(max_k):
                    current_theta = theta * k / (max_k - 1)
                    quat = glm.quat(math.cos(current_theta/2), math.sin(current_theta/2)*up)

                    H = first_hit_point - glm.dot(first_hit_point, up) * up # 原点到轴的垂足

                    for j in range(len_section):
                        point = current_section[j]
                        pos = quat * (point - H) + H

                        s = L + distance_first_hit_to_corner*current_theta
                        t = accumulate_length[j]
                        if normalize_tex_coord:
                            s /= accumulate_length[-1]
                            t /= accumulate_length[-1]

                        vertex = Vertex()
                        vertex.position = pos
                        vertex.tex_coord = glm.vec3(s, t, 0)

                        vertices[i_vertex] = vertex
                        i_vertex += 1

                        if i > 0 and j > 0:
                            triangle = glm.uvec3(0, 0, 0)
                            triangle[0] = i_vertex - 1
                            triangle[1] = i_vertex - 2 - len_section
                            triangle[2] = i_vertex - 2
                            indices[i_index] = triangle
                            i_index += 1
                            self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                            triangle = glm.uvec3(0, 0, 0)
                            triangle[0] = i_vertex - 1
                            triangle[1] = i_vertex - 1 - len_section
                            triangle[2] = i_vertex - 2 - len_section
                            indices[i_index] = triangle
                            i_index += 1
                            self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])
                    
                            yield

                current_section = backup_section
                L += distance_first_hit_to_corner*theta
            else:
                for j in range(len_section):
                    t = accumulate_length[j]
                    s = L + glm.dot(current_section[j]-path[i], d1)                    
                    if normalize_tex_coord:
                        t /= accumulate_length[-1]
                        s /= accumulate_length[-1]
                    
                    vertex = Vertex()
                    vertex.position = current_section[j]
                    vertex.tex_coord = glm.vec3(s, t, 0)
                    vertices[i_vertex] = vertex
                    i_vertex += 1

                    if i > 0 and j > 0:
                        triangle = glm.uvec3(0, 0, 0)
                        triangle[0] = i_vertex-1
                        triangle[1] = i_vertex-2-len_section
                        triangle[2] = i_vertex-2
                        indices[i_index] = triangle
                        i_index += 1
                        self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                        triangle = glm.uvec3(0, 0, 0)
                        triangle[0] = i_vertex-1
                        triangle[1] = i_vertex-1-len_section
                        triangle[2] = i_vertex-2-len_section
                        indices[i_index] = triangle
                        i_index += 1
                        self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                        yield

            last_should_add_corner = should_add_corner

        del vertices[i_vertex:]
        del indices[i_index:]

    def miter_join_build(self, current_section, accumulate_length, section_offset, path):
        vertices = self.vertices
        indices = self.indices
        normalize_tex_coord = self.__normalize_tex_coord
        len_section = len(current_section)

        i_vertex = 0
        i_index = 0

        L = 0
        d1, d2 = None, None
        len_path = len(path)
        for i in range(len_path):
            should_add_corner = True
            if i == 0:
                d1 = glm.normalize(path[i+1]-path[i])
                should_add_corner = False
            else:
                d1 = path[i] - path[i-1]
                len_d1 = glm.length(d1)
                L += len_d1
                d1 /= len_d1
                
                if i < len_path-1:
                    d2 = glm.normalize(path[i+1] - path[i])
                    half_way = glm.normalize(-d1 + d2)
                    up = glm.normalize(glm.cross(d1, d2))
                    cos_theta = glm.dot(d1, d2)

                    plane_normal = d1
                    if cos_theta > 0.9:
                        should_add_corner = False
                    else:
                        plane_normal = glm.normalize(glm.cross(half_way, up))

                    for j in range(len_section):
                        current_section[j] = line_intersect_plane(current_section[j], d1, path[i], plane_normal)
                else:
                    should_add_corner = False
                    for j in range(len_section):
                        current_section[j] = line_intersect_plane(current_section[j], d1, path[i], d1) + section_offset[j]*d1

            for j in range(len_section):
                t = accumulate_length[j]
                s = L + glm.dot(current_section[j] - path[i], d1)

                if normalize_tex_coord:
                    t /= accumulate_length[-1]
                    s /= accumulate_length[-1]

                vertex = Vertex()
                vertex.position = current_section[j]
                vertex.tex_coord = glm.vec3(s, t, 0)
                vertices[i_vertex] = vertex
                i_vertex += 1

                if i > 0 and j > 0:
                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 2 - len_section
                    triangle[2] = i_vertex - 2
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 1 - len_section
                    triangle[2] = i_vertex - 2 - len_section
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    yield

            if should_add_corner:
                for j in range(len_section):
                    t = accumulate_length[j]
                    s = L - glm.dot(current_section[j] - path[i], d1)
                    if normalize_tex_coord:
                        t /= accumulate_length[-1]
                        s /= accumulate_length[-1]

                    vertex = Vertex()
                    vertex.position = current_section[j]
                    vertex.tex_coord = glm.vec3(s, t, 0)
                    vertices[i_vertex] = vertex
                    i_vertex += 1

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def section(self):
        return self.__section
    
    @section.setter
    def section(self, section):
        self.__section = section
        self.start_building()

    @property
    def __used_section(self):
        should_duplicate_indices = set()
        for i in range(1, len(self.__section)-1):
            v1 = self.__section[i-1] - self.__section[i]
            v2 = self.__section[i+1] - self.__section[i]
            cos_angle = cos_angle_of(v1, v2)
            if cos_angle > -0.9:
                should_duplicate_indices.add(i)

        used_section = []
        for i in range(len(self.__section)):
            used_section.append(self.__section[i])
            if i in should_duplicate_indices:
                used_section.append(self.__section[i])

        return used_section

    @property
    def path(self):
        return self.__path
    
    @path.setter
    def path(self, path):
        self.__path = path
        self.start_building()

    @property
    def join_style(self):
        return self.__join_style
    
    @join_style.setter
    @Mesh.param_setter
    def join_style(self, join_style:JoinStyle):
        self.__join_style = join_style

    @property
    def n_corner_divide(self):
        return self.__n_corner_divide
    
    @n_corner_divide.setter
    @Mesh.param_setter
    def n_corner_divide(self, n:int):
        self.__n_corner_divide = n

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord

    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag


from glass import Vertex, Vertices, Indices, ShaderProgram, sampler2D, GLConfig, sampler2DArray
from glass.utils import modify_time, cat
from glass.GlassConfig import GlassConfig

import glm
import os
from OpenGL import GL

def _init_Frame(cls):
    cls.vertices = Vertices()
    cls.indices = Indices()

    cls.vertices[0] = Vertex(position=glm.vec2(-1, -1))
    cls.vertices[1] = Vertex(position=glm.vec2(1, -1))
    cls.vertices[2] = Vertex(position=glm.vec2(1, 1))
    cls.vertices[3] = Vertex(position=glm.vec2(-1, 1))

    cls.indices[0] = glm.uvec3(0, 1, 2)
    cls.indices[1] = glm.uvec3(0, 2, 3)

    self_folder = os.path.dirname(os.path.abspath(__file__)).replace("\\", "/")
    cls.draw_frame_vs = os.path.abspath(self_folder + "/glsl/Pipelines/draw_frame.vs")
    cls.draw_frame_fs = os.path.abspath(self_folder + "/glsl/Pipelines/draw_frame.fs")

    cls.program = ShaderProgram()
    cls.program.compile(cls.draw_frame_vs)
    cls.program.compile(cls.draw_frame_fs)
    cls.program.uniform_not_set_warning = False

    return cls

@_init_Frame
class Frame:

    __array_geo_shader_template_content = None

    @staticmethod
    def draw(screen_image:(sampler2D,sampler2DArray), gray:bool=False, invert:bool=False, layer:int=-1, index:int=0):
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            if isinstance(screen_image, sampler2D):
                Frame.program["screen_image"] = screen_image
                Frame.program["layer"] = -1
            else:
                Frame.program["screen_image_array"] = screen_image
                Frame.program["layer"] = layer

            Frame.program["gray"] = gray
            Frame.program["invert"] = invert
            Frame.program["index"] = index
            Frame.program.draw_triangles(Frame.vertices, Frame.indices)

    @staticmethod
    def draw_frame_array_gs(layers):
        self_folder = os.path.dirname(os.path.abspath(__file__))

        target_filename = GlassConfig.cache_folder + f"/draw_frame_array{layers}.gs"
        template_filename = self_folder + "/glsl/Pipelines/draw_frame_array.gs"
        if modify_time(template_filename) > modify_time(target_filename):
            if Frame.__array_geo_shader_template_content is None:
                Frame.__array_geo_shader_template_content = cat(template_filename)
            
            target_file = open(target_filename, "w")
            target_file.write(Frame.__array_geo_shader_template_content.replace("{layers}", str(layers)))
            target_file.close()

        return target_filename


import os

from OpenGL import GL
import re
import copy
import warnings
import sys

from .minifyc import minifyc
from .treeshake import treeshake, macros_expand
from .utils import di, defines_key, delete, md5s, modify_time, load_var, save_var, cat, relative_path, printable_path, printable_size
from .GlassConfig import GlassConfig
from .GLConfig import GLConfig
from .GLObject import GLObject
from .ShaderParser import ShaderParser
from .GPUProgram import CompileError, CompileWarning

class BaseShader(GLObject):

	__message_prefix1 = re.compile(r"^0\((?P<line_number>\d+)\) : (?P<message_type>\w+) ", flags=re.M)
	__message_prefix2 = re.compile(r"^(?P<message_type>\w+): 0:(?P<line_number>\d+): ", flags=re.M)
	__message_prefix3 = re.compile(r"^(?P<message_type>\w+): ([a-zA-Z]:[/\\]+)?[/\\]*([^<>:\"/\\|?*]+[/\\]+)*([^<>:\"/\\|?*]+)?:(?P<line_number>\d+): ", flags=re.M)
	_basic_info = \
	{
		"gen_func": GL.glCreateShader,
		"bind_func": None,
		"del_func": GL.glDeleteShader,
		"target_type": None,
		"binding_type": None,
		"need_number": False,
	}
	_type = None

	def __init__(self, program):
		GLObject.__init__(self)
		self._program_id:int = id(program)
		self._code:str = ""
		self._clean_code:str = ""
		self._comments_set:set = set()
		self._line_message_map:dict = {}
		self._file_name:str = ""
		self._compiled_but_not_applied:bool = False
		self._should_recompile:bool = True
		self._max_modify_time:int = 0

		self.attributes_info:dict = {}
		self.geometry_in:str = ""
		self.uniforms_info:dict = {}
		self.uniform_blocks_info:dict = {}
		self.shader_storage_blocks_info:dict = {}
		self.structs_info:dict = {}
		self.outs_info:dict = {}
		self.work_group_size:tuple = tuple()
		self.related_files:list = []

		self._include_paths:list = [""]
		self._defines:dict = {}

	@delete
	def clear(self):
		pass

	@delete
	def bind(self):
		pass

	@delete
	def unbind(self):
		pass

	@delete
	def is_bound(self):
		pass

	@property
	def type(self):
		return self._type

	@property
	def is_compiled(self):
		return bool(self._code)

	@property
	def file_name(self):
		return self._file_name

	@property
	def code(self):
		return self._code

	@property
	def clean_code(self):
		return self._clean_code

	def _apply(self):
		if not self._compiled_but_not_applied:
			return
		
		if self._id == 0:
			self._id = GL.glCreateShader(self._type)
			if self._id == 0:
				raise MemoryError("Failed to create Shader!")

		used_code = self._code
		if not GlassConfig.debug:
			# used_code = self._code
			# used_code = minifyc(self._code)
			used_code = minifyc(treeshake(self._code))

		if GlassConfig.print:
			print(f"compiling shader: {printable_path(self.file_name)} {printable_size(used_code)} ", end="", flush=True)

		if GlassConfig.debug:
			dest_folder = "used_shaders/" + os.path.basename(sys.argv[0])
			if not os.path.isdir(dest_folder):
				os.makedirs(dest_folder)

			out_file = open(dest_folder + "/" + os.path.basename(self.file_name), "w")
			out_file.write(self._code)
			out_file.close()

		GL.glShaderSource(self._id, used_code)
		GL.glCompileShader(self._id)

		message_bytes = GL.glGetShaderInfoLog(self._id)
		message = message_bytes
		if isinstance(message_bytes, bytes):
			message = str(message_bytes, encoding="utf-8")

		error_messages, warning_messages = self._format_error_warning(message)
		if warning_messages and GlassConfig.warning:
			warning_message = f"\nWarning when compiling: {self._file_name}:\n" + "\n".join(warning_messages)
			warnings.warn(warning_message, category=CompileWarning)

		if error_messages:
			error_message = f"\nError when compiling: {self._file_name}:\n" + "\n".join(error_messages)
			raise CompileError(error_message)
		
		success = GL.glGetShaderiv(self._id, GL.GL_COMPILE_STATUS)
		if success != GL.GL_TRUE:
			raise CompileError(message)
		
		meta_info = {}
		meta_info["related_files"] = self.related_files
		meta_info["code"] = self._code
		meta_info["clean_code"] = self._clean_code
		meta_info["line_message_map"] = self._line_message_map
		meta_info["attributes_info"] = self.attributes_info
		meta_info["geometry_in"] = self.geometry_in
		meta_info["uniforms_info"] = self.uniforms_info
		meta_info["uniform_blocks_info"] = self.uniform_blocks_info
		meta_info["shader_storage_blocks_info"] = self.shader_storage_blocks_info
		meta_info["structs_info"] = self.structs_info
		meta_info["outs_info"] = self.outs_info
		meta_info["work_group_size"] = self.work_group_size
		save_var(meta_info, self._meta_file_name)
		
		self._compiled_but_not_applied = False

		if GlassConfig.print:
			print("done")

	def _test_should_recompile(self):
		if GlassConfig.recompile:
			self._should_recompile = True
			return True

		meta_mtime = modify_time(self._meta_file_name)
		self._max_modify_time = 0

		if meta_mtime == 0:
			self._should_recompile = True
			return True
		
		meta_info = load_var(self._meta_file_name)
		related_files = meta_info["related_files"]
		should_recompile = False
		for file_name in related_files:			
			mtime = modify_time(file_name)
			if mtime == 0 or mtime > meta_mtime:
				should_recompile = True

			if mtime > self._max_modify_time:
				self._max_modify_time = mtime

		if should_recompile:
			self._should_recompile = True
			return True
		
		self._code = meta_info["code"]
		self._clean_code = meta_info["clean_code"]
		self._line_message_map = meta_info["line_message_map"]
		self.attributes_info = meta_info["attributes_info"]
		self.geometry_in = meta_info["geometry_in"]
		self.uniforms_info = meta_info["uniforms_info"]
		self.uniform_blocks_info = meta_info["uniform_blocks_info"]
		self.shader_storage_blocks_info = meta_info["shader_storage_blocks_info"]
		self.structs_info = meta_info["structs_info"]
		self.outs_info = meta_info["outs_info"]
		self.work_group_size = meta_info["work_group_size"]
		self.related_files = related_files

		self._should_recompile = False
		return False

	def _define_shader_type(self):
		if self._type == GL.GL_VERTEX_SHADER:
			self.define("VERTEX_SHADER")
		elif self._type == GL.GL_TESS_CONTROL_SHADER:
			self.define("TESS_CONTROL_SHADER")
		elif self._type == GL.GL_TESS_EVALUATION_SHADER:
			self.define("TESS_EVALUATION_SHADER")
		elif self._type == GL.GL_GEOMETRY_SHADER:
			self.define("GEOMETRY_SHADER")
		elif self._type == GL.GL_FRAGMENT_SHADER:
			self.define("FRAGMENT_SHADER")
		elif self._type == GL.GL_COMPUTE_SHADER:
			self.define("COMPUTE_SHADER")

	def _collect_info(self, file_name):
		abs_name = os.path.abspath(file_name).replace("\\", "/")
		self._code = cat(abs_name)
		self.related_files = [abs_name]

		self.add_include_path(".")
		self._define_shader_type()
		include_path = os.path.dirname(abs_name)
		if not os.path.isabs(file_name):
			include_path = relative_path(include_path)
		self.add_include_path(include_path)
		self.related_files.extend(self._replace_includes())

	def compile(self, file_name:str):
		if self.is_compiled and not self._compiled_but_not_applied:
			self.delete()

		if not os.path.isfile(file_name):
			raise FileNotFoundError(file_name)
		
		rel_name = relative_path(file_name)
		abs_name = os.path.abspath(file_name).replace("\\", "/")
		used_name = rel_name if len(rel_name) < len(abs_name) else abs_name
		self._file_name = used_name
		base_name = os.path.basename(abs_name)

		md5_key = f"{GLConfig.renderer}/{abs_name}{defines_key(self.defines)}"
		md5_value = md5s(md5_key)
		self._meta_file_name = GlassConfig.cache_folder + "/" + base_name + "_" + md5_value + ".meta"

		if self._test_should_recompile():
			self._collect_info(file_name)

			self._clean_code = macros_expand(self._code)
			self.attributes_info = {}
			self.geometry_in = ""
			self.uniforms_info = {}
			self.uniform_blocks_info = {}
			self.shader_storage_blocks_info = {}
			self.structs_info = {}
			self.outs_info = {}
			self.work_group_size = tuple()

			self.uniforms_info = ShaderParser.find_uniforms(self._clean_code)
			self.uniform_blocks_info = ShaderParser.find_uniform_blocks(self._clean_code)
			self.shader_storage_blocks_info = ShaderParser.find_shader_storage_blocks(self._clean_code)
			self.structs_info = ShaderParser.find_structs(self._clean_code)

			if self._type == GL.GL_VERTEX_SHADER:
				self.attributes_info = ShaderParser.find_attributes(self._clean_code)

			if self._type == GL.GL_GEOMETRY_SHADER:
				self.geometry_in = ShaderParser.find_geometry_in(self._clean_code)

			if self._type == GL.GL_FRAGMENT_SHADER:
				self.outs_info = ShaderParser.find_outs(self._clean_code)

			if self._type == GL.GL_COMPUTE_SHADER:
				self.work_group_size = ShaderParser.find_work_group_size(self._clean_code)

		self._compiled_but_not_applied = True

	def add_include_path(self, include_path:str):
		full_name = os.path.abspath(include_path).replace("\\", "/")
		if self._include_paths and self._include_paths[0] == full_name:
			return False

		self._include_paths.insert(0, full_name)
		return True

	def remove_include_path(self, include_path:str):
		full_name = os.path.abspath(include_path).replace("\\", "/")
		if full_name not in self._include_paths:
			return False

		while full_name in self._include_paths:
			self._include_paths.remove(full_name)

		return True

	@property
	def include_paths(self)->list:
		return self._include_paths + di(self._program_id).include_paths

	def define(self, name:str, value=None)->bool:
		if name in self._defines and self._defines[name] == value:
			return False
		
		self._defines[name] = value
		return True
	
	def undef(self, name:str)->bool:
		if name not in self._defines:
			return False
		
		del self._defines[name]
		return True
	
	@property
	def defines(self)->dict:
		defines = {}
		program = di(self._program_id)
		defines.update(program.defines)
		defines.update(self._defines)
		return defines

	def _find_comments(self):
		self._comments_set.clear()
		pos_start = 0
		pos_end = 0
		should_break = False
		len_code = len(self._code)
		while True:
			pos_start = self._code.find("/*", pos_end)
			if pos_start == -1:
				break

			pos_end = self._code.find("*/", pos_start+2)
			if pos_end == -1:
				pos_end = len_code - 2
				should_break = True
			pos_end += 2

			self._comments_set.update(set(range(pos_start, pos_end)))
			if should_break:
				break

		pos_start = 0
		pos_end = 0
		should_break = False
		while True:
			pos_start = self._code.find("//", pos_end)	
			if pos_start == -1:
				break

			pos_end = self._code.find("\n", pos_start+2)

			if pos_end == -1:
				pos_end = len_code
				should_break = True

			self._comments_set.update(set(range(pos_start, pos_end)))
			if should_break:
				break

	def _replace_includes(self):
		defines_str = ""
		for name, value in self.defines.items():
			if value is None:
				defines_str += f"#define {name}\n"
			else:
				defines_str += f"#define {name} {value}\n"
		lines_of_defines_str = ShaderParser.lines(defines_str)
		
		len_version = len("#version")
		self._find_comments()
		pos_version = 0
		while True:
			pos_version = self._code.find("#version", pos_version)
			if pos_version == -1:
				if defines_str:
					self._code = defines_str + self._code
				break
			elif pos_version in self._comments_set:
				pos_version += len_version
				continue
			else:
				pos_endl = self._code.find("\n", pos_version)
				if defines_str:
					if pos_endl == -1:
						self._code = self._code + "\n" + defines_str
					else:
						self._code = self._code[:pos_endl+1] + defines_str + self._code[pos_endl+1:]
				break
		line_num_version = ShaderParser.line_of(self._code, pos_version)

		pos_include_start = 0
		pos_filename_start = 0
		pos_include_end = 0
		pos_filename_end = 0
		n_lines = ShaderParser.lines(self._code)
		self._line_message_map[0] = self._file_name + ": {message_type}: "
		for i in range(1, n_lines+1):
			if i <= line_num_version:
				self._line_message_map[i] = self._file_name + ":" + str(i) + ": {message_type}: "
			elif i < line_num_version + lines_of_defines_str:
				self._line_message_map[i] = self._file_name + ":" + str(line_num_version) + ": {message_type}: "
			else:
				self._line_message_map[i] = self._file_name + ":" + str(i-lines_of_defines_str+1) + ": {message_type}: "

		included_files = set()
		should_find_comments = False
		len_include = len("#include")
		while True:
			if should_find_comments:
				self._find_comments()

			pos_include_start = 0
			while True:
				pos_include_start = self._code.find("#include", pos_include_start)
				if pos_include_start == -1:
					return included_files

				if pos_include_start in self._comments_set:
					pos_include_start += len_include
				else:
					break

			pos_filename_start = pos_include_start + len_include
			pos_filename_start = ShaderParser.skip_space(self._code, pos_filename_start)
			if self._code[pos_filename_start] != '<' and self._code[pos_filename_start] != '"':
				line_num = ShaderParser.line_of(self._code, pos_filename_start)
				raise CompileError("\n" + self._line_message_map[line_num].format(message_type="error") + '#include file name must be enveloped by <...> or "..."')
			
			start_char = self._code[pos_filename_start]
			end_char = ('>' if start_char == '<' else '"')
			pos_filename_start += 1
			pos_filename_start = ShaderParser.skip_space(self._code, pos_filename_start)
			pos_filename_end = self._code.find(end_char, pos_filename_start)
			if pos_filename_end == -1:
				line_num = ShaderParser.line_of(self._code, pos_filename_start)
				raise CompileError("\n" + self._line_message_map[line_num].format(message_type="error") + '#include file name must be enveloped by <...> or "..."')
			
			pos_include_end = pos_filename_end + 1
			pos_filename_end -= 1
			pos_filename_end = ShaderParser.skip_space_reverse(self._code, pos_filename_end)
			include_filename = self._code[pos_filename_start : pos_filename_end+1]
			found = False

			for include_path in self.include_paths:
				if include_path and not os.path.isdir(include_path):
					continue

				full_name = include_filename
				if include_path:
					full_name = include_path + "/" + include_filename

				if not os.path.isfile(full_name):
					continue

				rel_name = relative_path(full_name)
				abs_name = os.path.abspath(full_name).replace("\\", "/")
				used_name = rel_name if len(rel_name) < len(abs_name) else abs_name

				include_content = ""
				if abs_name not in included_files:
					include_content = cat(abs_name)
					included_files.add(abs_name)
				else:
					include_content = " " * (pos_include_end - pos_include_start)

				self._code = self._code[:pos_include_start] + include_content + self._code[pos_include_end:]
				if include_content:
					include_line_num = ShaderParser.line_of(self._code, pos_include_start)
					include_lines = ShaderParser.lines(include_content)
					include_line_end = include_line_num + include_lines
					n_lines = ShaderParser.lines(self._code)

					old_line_message_map = copy.deepcopy(self._line_message_map)
					for i in range(include_line_end, n_lines+1):
						self._line_message_map[i] = old_line_message_map[i - include_lines + 1]

					for i in range(include_line_num, include_line_end):
						self._line_message_map[i] = used_name + ":" + str(i-include_line_num+1) + ": {message_type}: "
				
					self.add_include_path(os.path.dirname(full_name))
					should_find_comments = True
				else:
					should_find_comments = False

				found = True
				break
				
			if not found:
				line_num = ShaderParser.line_of(self._code, pos_filename_start)
				raise CompileError("\n" + self._line_message_map[line_num].format(message_type="error") + f'File "{include_filename}" not exists.')

		return included_files

	def _format_error_warning(self, message):
		def _replace_message(match):
			message_type = match.group('message_type').lower()
			line_number = int(match.group('line_number'))
			return self._line_message_map[line_number].format(message_type=message_type)

		message = BaseShader.__message_prefix1.sub(_replace_message, message)
		message = BaseShader.__message_prefix2.sub(_replace_message, message)
		message = BaseShader.__message_prefix3.sub(_replace_message, message)
		message = message.replace("syntax error syntax error", "syntax error")
		message = message.strip(" \t\n\r")

		warning_messages = []
		error_messages = []
		last = ""

		lines = message.split("\n")
		for line in lines:
			line = line.strip(" \t\n\r")
			if not line:
				continue

			if "error" in line.lower():
				error_messages.append(line)
				last = "error"
			elif "warning" in line.lower():
				warning_messages.append(line)
				last = "warning"
			elif last == "error":
				error_messages.append(line)
			elif last == "warning":
				warning_messages.append(line)

		return error_messages, warning_messages

class VertexShader(BaseShader):
	_type = GL.GL_VERTEX_SHADER

class FragmentShader(BaseShader):
	_type = GL.GL_FRAGMENT_SHADER

class GeometryShader(BaseShader):
	_type = GL.GL_GEOMETRY_SHADER

class ComputeShader(BaseShader):
	_type = GL.GL_COMPUTE_SHADER

class TessControlShader(BaseShader):
	_type = GL.GL_TESS_CONTROL_SHADER

class TessEvaluationShader(BaseShader):
	_type = GL.GL_TESS_EVALUATION_SHADER



from OpenGL import GL
import numpy as np

from .GLInfo import GLInfo
from .GLObject import GLObject
from .GlassConfig import GlassConfig

class BO(GLObject):

	def __init__(self, context_shared=True)->None:
		GLObject.__init__(self, context_shared=context_shared)
		self._nbytes:int = 0
		self._draw_type:GLInfo.draw_types = GL.GL_STATIC_DRAW

	def delete(self)->None:
		GLObject.delete(self)
		self._nbytes = 0

	def bufferData(self, value_array, draw_type:GLInfo.draw_types=GL.GL_STATIC_DRAW)->None:
		if GlassConfig.debug and self.__class__.__name__ in ["FBO", "RBO"]:
			raise AttributeError("'" + self.__class__.__name__ + "' object has no attribute 'bufferData'")

		array_bytes = 0
		if isinstance(value_array, np.ndarray):
			array_bytes = value_array.nbytes
		elif isinstance(value_array, bytes):
			array_bytes = len(value_array)
		elif isinstance(value_array, bytearray):
			array_bytes = len(value_array)
			value_array = bytes(value_array)
		else:
			value_array = np.array(value_array)
			array_bytes = value_array.nbytes

		self.bind()
		GL.glBufferData(self.__class__._basic_info["target_type"], array_bytes, value_array, draw_type)
		self._nbytes = array_bytes
		self._draw_type = draw_type

	def malloc(self, nbytes:int, draw_type:GLInfo.draw_types=GL.GL_STATIC_DRAW)->None:
		if GlassConfig.debug:
			if self.__class__.__name__ in ["FBO", "RBO"]:
				raise AttributeError("'" + self.__class__.__name__ + "' object has no attribute 'malloc'")

		self.bind()
		GL.glBufferData(self.__class__._basic_info["target_type"], nbytes, None, draw_type)
		self._nbytes = nbytes
		self._draw_type = draw_type

	def memmove(self, old_start:int, nbytes:int, new_start:int)->None:
		if GlassConfig.debug:
			if old_start < 0:
				raise ValueError("source start position should be positive, " + str(old_start) + " is passed")
			elif old_start >= self.nbytes:
				raise ValueError("source start position is out of range, max position is " + str(self._nbytes-1), ", " + str(old_start) + " is passed")

			if new_start < 0:
				raise ValueError("dest start position should be positive, " + str(new_start) + " is passed")
			elif new_start >= self.nbytes:
				raise ValueError("dest start position is out of range, max position is " + str(self._nbytes-1), ", " + str(new_start) + " is passed")

			if nbytes < 0:
				raise ValueError("'nbytes' should be positive, " + str(nbytes) + " is passed")

			if old_start + nbytes > self._nbytes:
				raise ValueError("source buffer end position is out of range, max position is " + str(self._nbytes) +  ", " + str(old_start + nbytes) + " is applied.")

		if self._nbytes == 0 or nbytes == 0 or old_start == new_start:
			return

		temp_bo_id = GL.glGenBuffers(1)
		GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, temp_bo_id)
		GL.glBufferData(GL.GL_COPY_WRITE_BUFFER, nbytes, None, self.draw_type)

		GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, self._id)
		GL.glCopyBufferSubData(GL.GL_COPY_READ_BUFFER, GL.GL_COPY_WRITE_BUFFER, old_start, 0, nbytes)
		
		GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, 0)
		GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, 0)
		
		self.bufferSubData(old_start, nbytes, bytes(nbytes))
		real_delta = 0
		if old_start-new_start > 0:
			real_delta = old_start-new_start
			self.copy_to(new_start, real_delta, self, new_start+nbytes)
		else:
			real_delta = min(new_start+nbytes, self._nbytes)-(old_start+nbytes)
			self.copy_to(old_start+nbytes, real_delta, self, old_start)

		GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, self._id)
		GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, temp_bo_id)
		GL.glCopyBufferSubData(GL.GL_COPY_READ_BUFFER, GL.GL_COPY_WRITE_BUFFER, 0, new_start, nbytes)
		GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, 0)
		GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, 0)
		GL.glDeleteBuffers(1, np.array([temp_bo_id]))
		temp_bo_id = 0

	def bufferSubData(self, start:int, nbytes:int, value_array)->None:
		if GlassConfig.debug:
			if self.__class__.__name__ in ["FBO", "RBO"]:
				raise AttributeError("'" + self.__class__.__name__ + "' object has no attribute 'bufferSubData'")

		array_nbytes = 0
		if isinstance(value_array, np.ndarray):
			array_nbytes = value_array.nbytes
			value_array = value_array.tobytes()
		elif isinstance(value_array, bytes):
			array_nbytes = len(value_array)
		elif isinstance(value_array, bytearray):
			array_nbytes = len(value_array)
			value_array = bytes(value_array)
		else:
			value_array = np.array(value_array)
			array_nbytes = value_array.nbytes
			value_array = value_array.tobytes()

		if GlassConfig.debug:
			if start < 0:
				raise ValueError("Memory start position should be positive, " + str(start) + " is passed")
			
			if nbytes < 0:
				raise ValueError("'nbytes' should be positive, " + str(nbytes) + " is passed")

			if start >= self._nbytes:
				raise ValueError("Memory start position is out of range, max position is " + str(self._nbytes-1) + ", " + str(start) + " is passed.")

			if start + nbytes > self._nbytes:
				raise ValueError("Memory end position is out of range, max position is " + str(self._nbytes) +  ", " + str(start + nbytes) + " is applied.")

			if nbytes > array_nbytes:
				raise ValueError("Need copy data is over value_array's size. Max data size is " + str(value_array.nbytes) + " bytes, " + str(nbytes) + " bytes is needed.")

		if nbytes == 0:
			return

		self.bind()
		GL.glBufferSubData(self.__class__._basic_info["target_type"], start, nbytes, value_array)

	def copy_to(self, src_start:int, nbytes:int, dest_bo, dest_start:int)->None:
		if GlassConfig.debug:
			if self.__class__.__name__ in ["FBO", "RBO"]:
				raise AttributeError("'" + self.__class__.__name__ + "' object has no attribute 'copy_to'")

			if src_start < 0:
				raise ValueError("source start position should be positive, " + str(src_start) + " is passed")
			elif src_start >= self.nbytes:
				raise ValueError("source start position is out of range, max position is " + str(self._nbytes-1) + ", " + str(src_start) + " is passed")

			if dest_start < 0:
				raise ValueError("dest start position should be positive, " + str(dest_start) + " is passed")
			elif dest_start >= dest_bo.nbytes:
				raise ValueError("dest start position is out of range, max position is " + str(dest_bo._nbytes-1) + ", " + str(dest_start) + " is passed")
			
			if nbytes < 0:
				raise ValueError("'nbytes' should be positive, " + str(nbytes) + " is passed")

			if src_start + nbytes > self._nbytes:
				raise ValueError("source buffer end position is out of range, max position is " + str(self._nbytes) +  ", " + str(src_start + nbytes) + " is applied.")

		if self._nbytes == 0 or nbytes == 0:
			return

		real_nbytes = min(nbytes, dest_bo._nbytes-dest_start)
		if self._id != dest_bo._id:
			GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, self._id)
			GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, dest_bo._id)
			GL.glCopyBufferSubData(GL.GL_COPY_READ_BUFFER, GL.GL_COPY_WRITE_BUFFER, src_start, dest_start, real_nbytes)
			GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, 0)
			GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, 0)
		else:
			temp_bo_id = GL.glGenBuffers(1)
			GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, temp_bo_id)
			GL.glBufferData(GL.GL_COPY_WRITE_BUFFER, real_nbytes, None, self.draw_type)

			GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, self._id)
			GL.glCopyBufferSubData(GL.GL_COPY_READ_BUFFER, GL.GL_COPY_WRITE_BUFFER, src_start, 0, real_nbytes)

			GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, temp_bo_id)
			GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, self._id)
			GL.glCopyBufferSubData(GL.GL_COPY_READ_BUFFER, GL.GL_COPY_WRITE_BUFFER, 0, dest_start, real_nbytes)

			GL.glBindBuffer(GL.GL_COPY_READ_BUFFER, 0)
			GL.glBindBuffer(GL.GL_COPY_WRITE_BUFFER, 0)

			GL.glDeleteBuffers(1, np.array([temp_bo_id]))
			temp_bo_id = 0

	def shallow_copy_to(self, dest_bo)->None:
		self_members = dir(self)
		dest_members = dir(dest_bo)
		for key in self_members:
			if key in dest_members and not (key.startswith("__") and key.endswith("__")) and not callable(getattr(self, key)):
				setattr(dest_bo, key, getattr(self, key))

	@property
	def nbytes(self)->int:
		if GlassConfig.debug:
			if self.__class__.__name__ in ["FBO", "RBO"]:
				raise AttributeError("'" + self.__class__.__name__ + "' object has no attribute 'nbytes'")

		return self._nbytes

	@property
	def empty(self)->bool:
		return (self._nbytes == 0)

	@property
	def draw_type(self)->GLInfo.draw_types:
		if GlassConfig.debug:
			if self.__class__.__name__ in ["FBO", "RBO"]:
				raise AttributeError("'" + self.__class__.__name__ + "' object has no attribute 'draw_type'")
			
		return self._draw_type


from ..Mesh import Mesh
from ..ColorMap import ColorMap
from .Surf import Surf

from glass.utils import checktype

import numpy as np
import glm
import math

class SphericalFSurf(Mesh):

    @checktype
    def __init__(self, func, lon_range=[0,2*math.pi], lat_range=[-math.pi/2,math.pi/2],
                 color_map:ColorMap=None, back_color_map:ColorMap=None,
                 color:(glm.vec3,glm.vec4)=None, back_color:(glm.vec3,glm.vec4)=None,
                 surf_type:Mesh.SurfType=Mesh.SurfType.Smooth,
                 name:str=""):
        Mesh.__init__(self, name=name, surf_type=surf_type)
        self._func = func
        self._lon_range = lon_range
        self._lat_range = lat_range

        Surf._set_colors(self, color, back_color, color_map, back_color_map)
        self.start_building()
            
    def build(self):
        self.should_add_color = False

        func = self._func
        lon_range = self._lon_range
        lat_range = self._lat_range

        lon = np.linspace(lon_range[0], lon_range[1]) if len(lon_range) == 2 else lon_range
        lat = np.linspace(lat_range[0], lat_range[1]) if len(lat_range) == 2 else lat_range
        lon, lat = np.meshgrid(lon, lat)

        try:
            r = func(lon, lat)
            if r.shape != lon.shape:
                raise Exception()
        except:
            func = np.vectorize(func)
            r = func(lon, lat)

        X = r*np.cos(lat)*np.cos(lon)
        Y = r*np.cos(lat)*np.sin(lon)
        Z = r*np.sin(lat)
        C = r
        
        Surf._build(self, X, Y, Z, C, C)

    @property
    def func(self):
        return self._func
    
    @func.setter
    @Mesh.param_setter
    def func(self, func):
        self._func = func

    @property
    def lon_range(self):
        return self._lon_range
    
    @lon_range.setter
    @Mesh.param_setter
    def lon_range(self, range):
        self._lon_range = range

    @property
    def lat_range(self):
        return self._lat_range
    
    @lat_range.setter
    @Mesh.param_setter
    def lat_range(self, range):
        self._lat_range = range

    @property
    def color_map(self):
        return self._color_map
    
    @color_map.setter
    @Mesh.param_setter
    def color_map(self, color_map:ColorMap):
        self._color_map = color_map
        self._use_color_map = True
        if not self._back_color_map_user_set:
            self._back_color_map = color_map
            self._back_use_color_map = True

    @property
    def back_color_map(self):
        return self._back_color_map
    
    @back_color_map.setter
    @Mesh.param_setter
    def back_color_map(self, color_map:ColorMap):
        self._back_color_map = color_map
        self._back_use_color_map = True
        self._back_color_map_user_set = True



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm

class Box(Mesh):

    @checktype
    def __init__(self, Lx:float=1, Ly:float=None, Lz:float=None,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord=True,
                 name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        if Ly is None:
            Ly = Lx

        if Lz is None:
            Lz = Lx
        
        self.__Lx = Lx
        self.__Ly = Ly
        self.__Lz = Lz
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices

        Lx = self.__Lx
        Ly = self.__Ly
        Lz = self.__Lz
        normalize_tex_coord = self.__normalize_tex_coord

        # 左面
        position = glm.vec3(-Lx/2,  Ly/2, -Lz/2)
        tex_coord = glm.vec3(0) if normalize_tex_coord else glm.vec3(0.5-position.y, 0.5+position.z, 0)
        vertices[0] = Vertex(
            position=position,
            normal=glm.vec3(-1, 0, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2, -Ly/2, -Lz/2)
        tex_coord = glm.vec3(1, 0, 0) if normalize_tex_coord else glm.vec3(0.5-position.y, 0.5+position.z, 0)
        vertices[1] = Vertex(
            position=position,
            normal=glm.vec3(-1, 0, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2, -Ly/2,  Lz/2)
        tex_coord = glm.vec3(1, 1, 0) if normalize_tex_coord else glm.vec3(0.5-position.y, 0.5+position.z, 0)
        vertices[2] = Vertex(
            position=position,
            normal=glm.vec3(-1, 0, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2,  Ly/2,  Lz/2)
        tex_coord = glm.vec3(0, 1, 0) if normalize_tex_coord else glm.vec3(0.5-position.y, 0.5+position.z, 0)
        vertices[3] = Vertex(
            position=position,
            normal=glm.vec3(-1, 0, 0),
            tex_coord=tex_coord)

        # 右面
        position = glm.vec3(Lx/2, -Ly/2, -Lz/2)
        tex_coord = glm.vec3(0) if normalize_tex_coord else glm.vec3(0.5+position.y, 0.5+position.z, 0)
        vertices[4] = Vertex(
            position=position,
            normal=glm.vec3(1, 0, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(Lx/2,  Ly/2, -Lz/2)
        tex_coord = glm.vec3(1, 0, 0) if normalize_tex_coord else glm.vec3(0.5+position.y, 0.5+position.z, 0)
        vertices[5] = Vertex(
            position=position,
            normal=glm.vec3(1, 0, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(Lx/2,  Ly/2,  Lz/2)
        tex_coord = glm.vec3(1, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.y, 0.5+position.z, 0)
        vertices[6] = Vertex(
            position=position,
            normal=glm.vec3(1, 0, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(Lx/2, -Ly/2,  Lz/2)
        tex_coord = glm.vec3(0, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.y, 0.5+position.z, 0)
        vertices[7] = Vertex(
            position=position,
            normal=glm.vec3(1, 0, 0),
            tex_coord=tex_coord)

        # 后面
        position = glm.vec3(-Lx/2, -Ly/2, -Lz/2)
        tex_coord = glm.vec3(0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.z, 0)
        vertices[8] = Vertex(
            position=position,
            normal=glm.vec3(0, -1, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2, -Ly/2, -Lz/2)
        tex_coord = glm.vec3(1, 0, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.z, 0)
        vertices[9] = Vertex(
            position=position,
            normal=glm.vec3(0, -1, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2, -Ly/2,  Lz/2)
        tex_coord = glm.vec3(1, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.z, 0)
        vertices[10] = Vertex(
            position=position,
            normal=glm.vec3(0, -1, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2, -Ly/2,  Lz/2)
        tex_coord = glm.vec3(0, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.z, 0)
        vertices[11] = Vertex(
            position=position,
            normal=glm.vec3(0, -1, 0),
            tex_coord=tex_coord)

        # 前面
        position = glm.vec3( Lx/2, Ly/2, -Lz/2)
        tex_coord = glm.vec3(0) if normalize_tex_coord else glm.vec3(0.5-position.x, 0.5+position.z, 0)
        vertices[12] = Vertex(
            position=position,
            normal=glm.vec3(0, 1, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2, Ly/2, -Lz/2)
        tex_coord = glm.vec3(1, 0, 0) if normalize_tex_coord else glm.vec3(0.5-position.x, 0.5+position.z, 0)
        vertices[13] = Vertex(
            position=position,
            normal=glm.vec3(0, 1, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2, Ly/2,  Lz/2)
        tex_coord = glm.vec3(1, 1, 0) if normalize_tex_coord else glm.vec3(0.5-position.x, 0.5+position.z)
        vertices[14] = Vertex(
            position=position,
            normal=glm.vec3(0, 1, 0),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2, Ly/2,  Lz/2)
        tex_coord = glm.vec3(0, 1, 0) if normalize_tex_coord else glm.vec3(0.5-position.x, 0.5+position.z)
        vertices[15] = Vertex(
            position=position,
            normal=glm.vec3(0, 1, 0),
            tex_coord=tex_coord)

        # 下面
        position = glm.vec3(-Lx/2,  Ly/2, -Lz/2)
        tex_coord = glm.vec3(0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5-position.y)
        vertices[16] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, -1),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2,  Ly/2, -Lz/2)
        tex_coord = glm.vec3(1, 0, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5-position.y)
        vertices[17] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, -1),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2, -Ly/2, -Lz/2)
        tex_coord = glm.vec3(1, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5-position.y)
        vertices[18] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, -1),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2, -Ly/2, -Lz/2)
        tex_coord = glm.vec3(0, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5-position.y)
        vertices[19] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, -1),
            tex_coord=tex_coord)

        # 上面
        position = glm.vec3(-Lx/2, -Ly/2, Lz/2)
        tex_coord = glm.vec3(0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.y)
        vertices[20] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, 1),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2, -Ly/2, Lz/2)
        tex_coord = glm.vec3(1, 0, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.y, 0)
        vertices[21] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, 1),
            tex_coord=tex_coord)
        
        position = glm.vec3( Lx/2,  Ly/2, Lz/2)
        tex_coord = glm.vec3(1, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.y, 0)
        vertices[22] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, 1),
            tex_coord=tex_coord)
        
        position = glm.vec3(-Lx/2,  Ly/2, Lz/2)
        tex_coord = glm.vec3(0, 1, 0) if normalize_tex_coord else glm.vec3(0.5+position.x, 0.5+position.y, 0)
        vertices[23] = Vertex(
            position=position,
            normal=glm.vec3(0, 0, 1),
            tex_coord=tex_coord)

        # 左面
        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        # 右面
        indices[2] = 4+glm.uvec3(0, 1, 2)
        indices[3] = 4+glm.uvec3(0, 2, 3)

        # 后面
        indices[4] = 8+glm.uvec3(0, 1, 2)
        indices[5] = 8+glm.uvec3(0, 2, 3)

        # 前面
        indices[6] = 12+glm.uvec3(0, 1, 2)
        indices[7] = 12+glm.uvec3(0, 2, 3)
        
        # 下面
        indices[8] = 16+glm.uvec3(0, 1, 2)
        indices[9] = 16+glm.uvec3(0, 2, 3)

        # 上面
        indices[10] = 20+glm.uvec3(0, 1, 2)
        indices[11] = 20+glm.uvec3(0, 2, 3)

        del vertices[24:]
        del indices[12:]

        for index in indices:
            self.generate_temp_TBN(vertices[index[0]], vertices[index[1]], vertices[index[2]])

    @property
    def Lx(self):
        return self.__Lx
    
    @Lx.setter
    @Mesh.param_setter
    def Lx(self, Lx:int):
        self.__Lx = Lx

    @property
    def Ly(self):
        return self.__Ly
    
    @Ly.setter
    @Mesh.param_setter
    def Ly(self, Ly:int):
        self.__Ly = Ly

    @property
    def Lz(self):
        return self.__Lz
    
    @Lz.setter
    @Mesh.param_setter
    def Lz(self, Lz:int):
        self.__Lz = Lz

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag
        


from .Mesh import Mesh
from .Camera import Camera
from .GlassEngineConfig import GlassEngineConfig

from glass.utils import checktype
from glass import Vertex, samplerCube, ShaderProgram

import glm
from OpenGL import GL
import os

class SkyBox(Mesh):

    @checktype
    def __init__(self, name:str=""):
        Mesh.__init__(self, name=name, block=True)
        self.render_hints.depth_func = "<="
        self.render_hints.cull_face = GL.GL_FRONT
        self.__skybox_map = None
        self.__program = None
        self.should_add_color = False
        self.start_building()

    @property
    def skybox_map(self):
        return self.__skybox_map
    
    @skybox_map.setter
    @checktype
    def skybox_map(self, skybox_map:samplerCube):
        self.__skybox_map = skybox_map
    
    @property
    def is_completed(self):
        return (self.__skybox_map is not None and self.__skybox_map.is_completed)
    
    def __getattr__(self, key):
        if key not in ["right", "left", "top", "bottom", "back", "front", "up", "down"]:
            return super().__getattr__(key)

        if self.skybox_map is None:
            self.skybox_map = samplerCube()

        return self.skybox_map[key]

    def __setattr__(self, key, value):
        if key not in ["right", "left", "top", "bottom", "back", "front", "up", "down"]:
            return super().__setattr__(key, value)

        if self.skybox_map is None:
            self.skybox_map = samplerCube()

        self.skybox_map[key] = value

    def build(self):
        self.self_calculated_normal = True
        self.is_closed = True
        self.x_min, self.x_max = -1, 1
        self.y_min, self.y_max = -1, 1
        self.z_min, self.z_max = -1, 1

        vertices = self.vertices
        indices = self.indices

        # 左面
        vertices[0] = Vertex(position=glm.vec3(-1,  1, -1))
        vertices[1] = Vertex(position=glm.vec3(-1, -1, -1))
        vertices[2] = Vertex(position=glm.vec3(-1, -1,  1))
        vertices[3] = Vertex(position=glm.vec3(-1,  1,  1))

        # 右面
        vertices[4] = Vertex(position=glm.vec3(1, -1, -1))
        vertices[5] = Vertex(position=glm.vec3(1,  1, -1))
        vertices[6] = Vertex(position=glm.vec3(1,  1,  1))
        vertices[7] = Vertex(position=glm.vec3(1, -1,  1))

        # 后面
        vertices[8] = Vertex(position=glm.vec3(-1, -1, -1))
        vertices[9] = Vertex(position=glm.vec3( 1, -1, -1))
        vertices[10] = Vertex(position=glm.vec3( 1, -1,  1))
        vertices[11] = Vertex(position=glm.vec3(-1, -1,  1))

        # 前面
        vertices[12] = Vertex(position=glm.vec3( 1, 1, -1))
        vertices[13] = Vertex(position=glm.vec3(-1, 1, -1))
        vertices[14] = Vertex(position=glm.vec3(-1, 1,  1))
        vertices[15] = Vertex(position=glm.vec3( 1, 1,  1))

        # 下面
        vertices[16] = Vertex(position=glm.vec3(-1,  1, -1))
        vertices[17] = Vertex(position=glm.vec3( 1,  1, -1))
        vertices[18] = Vertex(position=glm.vec3( 1, -1, -1))
        vertices[19] = Vertex(position=glm.vec3(-1, -1, -1))

        # 上面
        vertices[20] = Vertex(position=glm.vec3(-1, -1, 1))
        vertices[21] = Vertex(position=glm.vec3( 1, -1, 1))
        vertices[22] = Vertex(position=glm.vec3( 1,  1, 1))
        vertices[23] = Vertex(position=glm.vec3(-1,  1, 1))

        # 左面
        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        # 右面
        indices[2] = 4+glm.uvec3(0, 1, 2)
        indices[3] = 4+glm.uvec3(0, 2, 3)

        # 后面
        indices[4] = 8+glm.uvec3(0, 1, 2)
        indices[5] = 8+glm.uvec3(0, 2, 3)

        # 前面
        indices[6] = 12+glm.uvec3(0, 1, 2)
        indices[7] = 12+glm.uvec3(0, 2, 3)
        
        # 下面
        indices[8] = 16+glm.uvec3(0, 1, 2)
        indices[9] = 16+glm.uvec3(0, 2, 3)

        # 上面
        indices[10] = 20+glm.uvec3(0, 1, 2)
        indices[11] = 20+glm.uvec3(0, 2, 3)

    @property
    def program(self):
        if self.__program is None:
            self.__program = ShaderProgram()
            self_folder = os.path.dirname(os.path.abspath(__file__))
            self.__program.compile(self_folder + "/glsl/Pipelines/skybox/skybox.vs")
            self.__program.compile(self_folder + "/glsl/Pipelines/skybox/skybox.fs")

        return self.__program
    
    def draw(self, camera:Camera):
        scene = camera.scene
        self.program["camera"] = camera
        self.program["skybox_map"] = self.skybox_map
        self.program["sky_distance"] = scene.background.distance
        if GlassEngineConfig["USE_FOG"]:
            self.program["fog"] = scene.fog
            
        Mesh.draw(self, self.program)


from glass.utils import checktype
from glass import sampler2D
from glass.ImageLoader import ImageLoader
from glass.WeakSet import WeakSet
from glass.MetaInstancesRecorder import MetaInstancesRecorder
from .callback_vec import callback_vec3, callback_vec4
from .GlassEngineConfig import GlassEngineConfig

import glm
from enum import Enum
import numpy as np
import math
from functools import wraps

def param_setter(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        value = args[1]

        should_test_transparent = False
        if func.__name__ in \
            ["diffuse_map", "ambient_map", "specular_map", "emission_map",
            "reflection_map", "refractive_index", "base_color_map"] and \
            not self._opacity_user_set and self._opacity == 0:
            self._opacity = 1
            should_test_transparent = True

            for mesh in self._parent_meshes:
                colors = None
                if mesh.material is self:
                    if "color" in mesh.vertices:
                        colors = mesh.vertices["color"].ndarray
                elif mesh._back_material is self:
                    if "back_color" in mesh.vertices:
                        colors = mesh.vertices["back_color"].ndarray

                if colors is not None and (len(colors.shape) != 2 or colors.shape[1] != 4):
                    colors = colors.reshape(-1, 4)
                
                if colors is not None and colors.size > 0 and np.all(colors == colors[0, :]):
                    used_color = glm.vec3(colors[0, 0], colors[0, 1], colors[0, 2])
                    
                    old_should_callback = self._should_callback
                    self._should_callback = False
                    self.diffuse = used_color
                    self.base_color = used_color
                    self._should_callback = old_should_callback

                break
        
        equal = False
        try:
            lvalue = getattr(self, func.__name__)
            if type(lvalue) != type(value):
                equal = False
            else:
                equal = bool(getattr(self, func.__name__) == value)
        except:
            equal = False

        if equal:
            return

        safe_func = checktype(func)
        return_value = safe_func(*args, **kwargs)

        self._update_all_env_maps()
        if func.__name__ == "generate_shadows":
            self._update_all_depth_maps()

        if should_test_transparent:
            self._test_transparent()

        return return_value

    return wrapper

class Material(metaclass=MetaInstancesRecorder):

    class ShadingModel(Enum):
        Flat = 0x1
        Gouraud = 0x2
        Phong = 0x3
        PhongBlinn = 0x4
        Toon = 0x5
        OrenNayar = 0x6
        Minnaert = 0x7
        CookTorrance = 0x8
        NoShading = 0x9
        Unlit = 0x9
        Fresnel = 0xa
        PBR = 0xb

    class Type(Enum):
        Emerald = "翡翠"
        Jade = "玉"
        Obsidian = "黑曜石"
        Pearl = "珍珠"
        Ruby = "红宝石"
        Turquoise = "绿松石"
        Brass = "黄铜"
        Bronze = "青铜"
        Chrome = "铬"
        Copper = "铜"
        Gold = "金"
        Silver = "银"
        BlackPlastic = "黑色塑料"
        CyanPlastic = "青色塑料"
        GreenPlastic = "绿色塑料"
        RedPlastic = "红色塑料"
        WhitePlastic = "白色塑料"
        YellowPlastic = "黄色塑料"
        BlackRubber = "黑色橡胶"
        CyanRubber = "青色橡胶"
        GreenRubber = "绿色橡胶"
        RedRubber = "红色橡胶"
        WhiteRubber = "白色橡胶"
        YellowRubber = "黄色橡胶"
        
        def __init__(self, ch_name):
            self.ch_name = ch_name

    @MetaInstancesRecorder.init
    def __init__(self, name:str=""):
        self._name:str = name
        self._shading_model:Material.ShadingModel = Material.ShadingModel.PhongBlinn

        self._ambient:callback_vec3 = callback_vec3(0, 0, 0, self._color_change_callback)
        self._diffuse:callback_vec3 = callback_vec3(0.396, 0.74151, 0.69102, self._color_change_callback)
        self._specular:callback_vec3 = callback_vec3(0.3, 0.3, 0.3, self._color_change_callback)
        self._shininess:float = 0.6*128
        self._shininess_strength:float = 1
        self._emission:callback_vec3 = callback_vec3(0, 0, 0, self._color_change_callback)
        self._emission_strength:float = 1
        self._opacity:float = 0
        self._height_scale:float = 0.05
        self._base_color:callback_vec3 = callback_vec3(0.5, 0.5, 0.5, self._color_change_callback)
        self._metallic:float = 0.5
        self._roughness:float = 0
        self._recv_shadows:bool = True
        self._cast_shadows:bool = True
        self._diffuse_bands:int = 2
        self._specular_bands:int = 2
        self._diffuse_softness:float = 0.05
        self._specular_softness:float = 0.02
        self._rim_power:float = 0.3
        self._fog:bool = True
        self._reflection:callback_vec4 = callback_vec4(1, 1, 1, 0, self._color_change_callback)
        self._refractive_index:float = 0
        self._env_mix_diffuse:bool = True
        self._env_max_bake_times:int = 2
        self._dynamic_env_mapping:bool = False
        self._auto_update_env_map:bool = False

        self._ambient_map:sampler2D = None
        self._diffuse_map:sampler2D = None
        self._specular_map:sampler2D = None
        self._shininess_map:sampler2D = None
        self._glossiness_map:sampler2D = None
        self._emission_map:sampler2D = None
        self._normal_map:sampler2D = None
        self._height_map:sampler2D = None
        self._opacity_map:sampler2D = None
        self._ao_map:sampler2D = None
        self._reflection_map:sampler2D = None
        self._base_color_map:sampler2D = None
        self._metallic_map:sampler2D = None
        self._roughness_map:sampler2D = None
        self._arm_map:sampler2D = None

        self._opacity_user_set:bool = False
        self._reflection_user_set:bool = False
        self._has_transparent:bool = True
        self._has_opaque:bool = False
        self._parent_meshes:WeakSet = WeakSet()
        self._should_callback:bool = True
        self._prop_name:str = ""

        GlassEngineConfig._update_cast_shadows(self._cast_shadows)
        GlassEngineConfig._update_recv_shadows(self._recv_shadows)
        GlassEngineConfig._update_recv_fog(self._fog)
        GlassEngineConfig._update_dynamic_env(self._dynamic_env_mapping)
        GlassEngineConfig._update_shading_model()

    @MetaInstancesRecorder.delete
    def __del__(self):
        pass

    @property
    def name(self)->str:
        return self._name
    
    @name.setter
    def name(self, name:str):
        self._name = name

    def _color_change_callback(self):
        if not self._should_callback:
            return

        old_should_callback = self._should_callback
        self._should_callback = False

        if self._prop_name == "reflection":
            self._reflection_user_set = True

        if self._prop_name == "ambient":
            if glm.length(self._ambient) < 1E-6:
                self.ambient = glm.vec3(1E-6)

        if not self._opacity_user_set and self._opacity == 0:
            self._opacity = 1
            for mesh in self._parent_meshes:
                colors = None
                if mesh.material is self:
                    if "color" in mesh.vertices:
                        colors = mesh.vertices["color"].ndarray
                elif mesh._back_material is self:
                    if "back_color" in mesh.vertices:
                        colors = mesh.vertices["back_color"].ndarray

                if colors is not None and (len(colors.shape) != 2 or colors.shape[1] != 4):
                    colors = colors.reshape(-1, 4)

                if colors is not None and colors.size > 0 and np.all(colors == colors[0, :]):
                    used_color = glm.vec3(colors[0, 0], colors[0, 1], colors[0, 2])
                    if self._prop_name != "diffuse":
                        self.diffuse = used_color

                    if self._prop_name != "base_color":
                        self.base_color = used_color

                break
                        
        self._update_all_env_maps()
        self._test_transparent()

        self._should_callback = old_should_callback

    @property
    def fog(self)->bool:
        return self._fog
    
    @fog.setter
    @param_setter
    def fog(self, fog:bool):
        self._fog = fog
        GlassEngineConfig._update_recv_fog(fog)

    @property
    def diffuse_bands(self)->int:
        return self._diffuse_bands
    
    @diffuse_bands.setter
    @param_setter
    def diffuse_bands(self, bands:int):
        self._diffuse_bands = bands

    @property
    def specular_bands(self)->int:
        return self._specular_bands
    
    @specular_bands.setter
    @param_setter
    def specular_bands(self, bands:int):
        self._specular_bands = bands

    @property
    def diffuse_softness(self)->float:
        return self._diffuse_softness
    
    @diffuse_softness.setter
    @param_setter
    def diffuse_softness(self, softness:float):
        self._diffuse_softness = max(softness, 1E-5)

    @property
    def specular_softness(self)->float:
        return self._specular_softness
    
    @specular_softness.setter
    @param_setter
    def specular_softness(self, softness:float):
        self._specular_softness = max(softness, 1E-5)

    @property
    def rim_power(self)->float:
        return self._rim_power
    
    @rim_power.setter
    @param_setter
    def rim_power(self, p:float):
        self._rim_power = p

    @property
    def recv_shadows(self):
        return self._recv_shadows
    
    @recv_shadows.setter
    @param_setter
    def recv_shadows(self, flag:bool):
        self._recv_shadows = flag
        GlassEngineConfig._update_recv_shadows(flag)

    @property
    def cast_shadows(self):
        return self._cast_shadows
    
    @cast_shadows.setter
    @param_setter
    def cast_shadows(self, flag:bool):
        self._cast_shadows = flag
        GlassEngineConfig._update_cast_shadows(flag)

    @property
    def has_transparent(self):
        return self._has_transparent
    
    @property
    def has_opaque(self):
        return self._has_opaque

    @property
    def need_env_map(self):
        if self._env_max_bake_times <= 0 or not self._dynamic_env_mapping:
            return False

        has_reflection = False
        if self._reflection_map is None:
            has_reflection = (glm.length(self._reflection.rgb*self._reflection.a) > 1E-6)
        else:
            has_reflection = True

        if has_reflection:
            return True
    
    @property
    def auto_update_env_map(self):
        return self._auto_update_env_map
    
    @auto_update_env_map.setter
    @param_setter
    def auto_update_env_map(self, flag:bool):
        self._auto_update_env_map = flag
    
    def update_env_map(self):
        for mesh in self._parent_meshes:
            for scene in mesh.scenes:
                for instance in scene._all_meshes[mesh].values():
                    instance.user_data["env_bake_times", scene] = 0

    @property
    def dynamic_env_mapping(self):
        return self._dynamic_env_mapping
    
    @dynamic_env_mapping.setter
    @param_setter
    def dynamic_env_mapping(self, flag:bool):
        self._dynamic_env_mapping = flag
        GlassEngineConfig._update_dynamic_env(flag)

    @property
    def env_max_bake_times(self):
        return self._env_max_bake_times
    
    @env_max_bake_times.setter
    @param_setter
    def env_max_bake_times(self, times:int):
        self._env_max_bake_times = times

    @property
    def shading_model(self):
        return self._shading_model
    
    @shading_model.setter
    @param_setter
    def shading_model(self, shading_model:(ShadingModel,None)):
        if shading_model is None:
            shading_model = Material.ShadingModel.Unlit

        self._shading_model = shading_model
        GlassEngineConfig._update_shading_model()

    @property
    def ambient(self)->callback_vec3:
        self._prop_name = "ambient"

        return self._ambient
    
    @ambient.setter
    def ambient(self, ambient:glm.vec3)->None:
        self._prop_name = "ambient"

        if glm.length(ambient) < 1E-6:
            ambient = glm.vec3(1E-6)

        old_should_callback = self._should_callback
        self._should_callback = False
        self._ambient.r = ambient.r
        self._ambient.g = ambient.g
        self._should_callback = old_should_callback

        self._ambient.b = ambient.b

    @property
    def diffuse(self)->callback_vec3:
        self._prop_name = "diffuse"

        return self._diffuse
    
    @diffuse.setter
    def diffuse(self, diffuse:glm.vec3)->None:
        self._prop_name = "diffuse"

        old_should_callback = self._should_callback
        self._should_callback = False
        self._diffuse.r = diffuse.r
        self._diffuse.g = diffuse.g
        self._should_callback = old_should_callback

        self._diffuse.b = diffuse.b

    @property
    def specular(self)->callback_vec3:
        self._prop_name = "specular"

        return self._specular
    
    @specular.setter
    def specular(self, specular:glm.vec3)->None:
        self._prop_name = "specular"

        old_should_callback = self._should_callback
        self._should_callback = False
        self._specular.r = specular.r
        self._specular.g = specular.g
        self._should_callback = old_should_callback

        self._specular.b = specular.b

    @property
    def glossiness(self):
        return math.sqrt(self._shininess)
    
    @glossiness.setter
    @param_setter
    def glossiness(self, glossiness:float):
        self._shininess = glossiness * glossiness

    @property
    def shininess(self):
        return self._shininess
    
    @shininess.setter
    @param_setter
    def shininess(self, shininess:float):
        self._shininess = shininess

    @property
    def emission_strength(self)->float:
        return self._emission_strength
    
    @emission_strength.setter
    @param_setter
    def emission_strength(self, emission_strength:float)->None:
        self._emission_strength = emission_strength

    @property
    def shininess_strength(self):
        return self._shininess_strength
    
    @shininess_strength.setter
    @param_setter
    def shininess_strength(self, shininess_strength:float):
        self._shininess_strength = shininess_strength

    @property
    def opacity(self):
        return self._opacity
    
    @opacity.setter
    @param_setter
    def opacity(self, opacity:float):
        self._opacity = opacity
        self._opacity_user_set = True

        self._test_transparent()

    @property
    def emission(self)->callback_vec3:
        self._prop_name = "emission"

        return self._emission
    
    @emission.setter
    def emission(self, emission:glm.vec3)->None:
        self._prop_name = "emission"

        old_should_callback = self._should_callback
        self._should_callback = False
        self._emission.r = emission.r
        self._emission.g = emission.g
        self._should_callback = old_should_callback

        self._emission.b = emission.b

    @property
    def env_mix_diffuse(self):
        return self._env_mix_diffuse
    
    @env_mix_diffuse.setter
    @param_setter
    def env_mix_diffuse(self, flag:bool):
        self._env_mix_diffuse = flag

    @property
    def reflection(self)->callback_vec4:
        self._prop_name = "reflection"
        return self._reflection
    
    @reflection.setter
    def reflection(self, reflection:(glm.vec4,glm.vec3,float))->None:
        if isinstance(reflection, glm.vec3):
            reflection = glm.vec4(reflection, 1)
        if isinstance(reflection, (float,int)):
            reflection = glm.vec4(1,1,1,reflection)

        self._prop_name = "reflection"
        self._reflection_user_set = True

        old_should_callback = self._should_callback
        self._should_callback = False
        self._reflection.r = reflection.r
        self._reflection.g = reflection.g
        self._reflection.b = reflection.b
        self._should_callback = old_should_callback

        self._reflection.a = reflection.a

    @property
    def refractive_index(self):
        return self._refractive_index
    
    @refractive_index.setter
    @param_setter
    def refractive_index(self, refractive_index:float):
        self._refractive_index = refractive_index

        if not self._reflection_user_set:
            self._reflection = glm.vec4(1, 1, 1, 1)

    @property
    def height_scale(self):
        return self._height_scale
    
    @height_scale.setter
    @param_setter
    def height_scale(self, distance:float):
        self._height_scale = distance

    @property
    def base_color(self)->callback_vec3:
        self._prop_name = "base_color"
        return self._base_color
    
    @base_color.setter
    def base_color(self, base_color:glm.vec3)->None:
        self._prop_name = "base_color"

        old_should_callback = self._should_callback
        self._should_callback = False
        self._base_color.r = base_color.r
        self._base_color.g = base_color.g
        self._should_callback = old_should_callback
        
        self._base_color.b = base_color.b

    @property
    def roughness(self):
        return self._roughness
    
    @roughness.setter
    @param_setter
    def roughness(self, roughness:float):
        self._roughness = roughness

    @property
    def metallic(self):
        return self._metallic
    
    @metallic.setter
    @param_setter
    def metallic(self, metallic:float):
        self._metallic = metallic

    @property
    def ambient_map(self):
        return self._ambient_map
    
    @ambient_map.setter
    @param_setter
    def ambient_map(self, ambient_map:(sampler2D,str,np.ndarray)):
        if isinstance(ambient_map, sampler2D) or ambient_map is None:
            self._ambient_map = ambient_map
        elif isinstance(ambient_map, (str,np.ndarray)):
            if isinstance(ambient_map, str):
                ambient_map = ImageLoader.load(ambient_map)
            threshold = 0.5
            image_dtype = ambient_map.dtype
            if "int" in str(image_dtype):
                threshold = 127
            if ambient_map.max() > threshold:
                ambient_map = (0.2 * ambient_map).astype(image_dtype)
            self._ambient_map = sampler2D(ambient_map)

    @property
    def diffuse_map(self):
        return self._diffuse_map
    
    @diffuse_map.setter
    @param_setter
    def diffuse_map(self, diffuse_map:(sampler2D,str,np.ndarray)):
        if isinstance(diffuse_map, sampler2D) or diffuse_map is None:
            self._diffuse_map = diffuse_map
        elif isinstance(diffuse_map, (str,np.ndarray)):
            if self._diffuse_map is None:
                self._diffuse_map = sampler2D(diffuse_map)
            else:
                self._diffuse_map.image = diffuse_map

        self._test_transparent()

    @property
    def specular_map(self):
        return self._specular_map
    
    @specular_map.setter
    @param_setter
    def specular_map(self, specular_map:(sampler2D,str,np.ndarray)):
        if isinstance(specular_map, sampler2D) or specular_map is None:
            self._specular_map = specular_map
        elif isinstance(specular_map, (str,np.ndarray)):
            if self._specular_map is None:
                self._specular_map = sampler2D(specular_map)
            else:
                self._specular_map.image = specular_map

    @property
    def shininess_map(self):
        return self._shininess_map
    
    @shininess_map.setter
    @param_setter
    def shininess_map(self, shininess_map:(sampler2D,str,np.ndarray)):
        if isinstance(shininess_map, sampler2D) or shininess_map is None:
            self._shininess_map = shininess_map
        elif isinstance(shininess_map, (str,np.ndarray)):
            if self._shininess_map is None:
                self._shininess_map = sampler2D(shininess_map)
            else:
                self._shininess_map.image = shininess_map

    @property
    def glossiness_map(self):
        return self._glossiness_map
    
    @glossiness_map.setter
    @param_setter
    def glossiness_map(self, glossiness_map:(sampler2D,str,np.ndarray)):
        if isinstance(glossiness_map, sampler2D) or glossiness_map is None:
            self._glossiness_map = glossiness_map
        elif isinstance(glossiness_map, (str,np.ndarray)):
            if self._glossiness_map is None:
                self._glossiness_map = sampler2D(glossiness_map)
            else:
                self._glossiness_map.image = glossiness_map

    @property
    def emission_map(self):
        return self._emission_map
    
    @emission_map.setter
    @param_setter
    def emission_map(self, emission_map:(sampler2D,str,np.ndarray)):
        if isinstance(emission_map, sampler2D) or emission_map is None:
            self._emission_map = emission_map
        elif isinstance(emission_map, (str,np.ndarray)):
            if self._emission_map is None:
                self._emission_map = sampler2D(emission_map)
            else:
                self._emission_map.image = emission_map

    @property
    def normal_map(self):
        return self._normal_map

    @normal_map.setter
    @param_setter
    def normal_map(self, normal_map:(sampler2D,str,np.ndarray)):
        if isinstance(normal_map, sampler2D) or normal_map is None:
            self._normal_map = normal_map
        elif isinstance(normal_map, (str,np.ndarray)):
            if self._normal_map is None:
                self._normal_map = sampler2D(normal_map)
            else:
                self._normal_map.image = normal_map

    @property
    def height_map(self):
        return self._height_map
    
    @height_map.setter
    @param_setter
    def height_map(self, height_map:(sampler2D,str,np.ndarray)):
        if isinstance(height_map, sampler2D) or height_map is None:
            self._height_map = height_map
        elif isinstance(height_map, (str,np.ndarray)):
            if self._height_map is None:
                self._height_map = sampler2D(height_map)
            else:
                self._height_map.image = height_map

    @property
    def opacity_map(self):
        return self._opacity_map
    
    @opacity_map.setter
    @param_setter
    def opacity_map(self, opacity_map:(sampler2D,str,np.ndarray)):
        if isinstance(opacity_map, sampler2D) or opacity_map is None:
            self._opacity_map = opacity_map
        elif isinstance(opacity_map, (str,np.ndarray)):
            if self._opacity_map is None:
                self._opacity_map = sampler2D(opacity_map)
            else:
                self._opacity_map.image = opacity_map

        self._opacity_user_set = True
        self._test_transparent()

    @property
    def ao_map(self):
        return self._ao_map
    
    @ao_map.setter
    @param_setter
    def ao_map(self, ao_map:(sampler2D,str,np.ndarray)):
        if isinstance(ao_map, sampler2D) or ao_map is None:
            self._ao_map = ao_map
        elif isinstance(ao_map, (str,np.ndarray)):
            if self._ao_map is None:
                self._ao_map = sampler2D(ao_map)
            else:
                self._ao_map.image = ao_map

    @property
    def arm_map(self):
        return self._arm_map
    
    @arm_map.setter
    @param_setter
    def arm_map(self, arm_map:(sampler2D,str,np.ndarray)):
        if isinstance(arm_map, sampler2D) or arm_map is None:
            self._arm_map = arm_map
        elif isinstance(arm_map, (str,np.ndarray)):
            if self._arm_map is None:
                self._arm_map = sampler2D(arm_map)
            else:
                self._arm_map.image = arm_map

    @property
    def reflection_map(self):
        return self._reflection_map
    
    @reflection_map.setter
    @param_setter
    def reflection_map(self, reflection_map:(sampler2D,str,np.ndarray)):
        if isinstance(reflection_map, sampler2D) or reflection_map is None:
            self._reflection_map = reflection_map
        elif isinstance(reflection_map, (str,np.ndarray)):
            if self._reflection_map is None:
                self._reflection_map = sampler2D(reflection_map)
            else:
                self._reflection_map.image = reflection_map

        self._reflection_user_set = True

    @property
    def base_color_map(self):
        return self._base_color_map
    
    @base_color_map.setter
    @param_setter
    def base_color_map(self, base_color_map:(sampler2D,str,np.ndarray)):
        if isinstance(base_color_map, sampler2D) or base_color_map is None:
            self._base_color_map = base_color_map
        elif isinstance(base_color_map, (str,np.ndarray)):
            if self._base_color_map is None:
                self._base_color_map = sampler2D(base_color_map)
            else:
                self._base_color_map.image = base_color_map

    @property
    def metallic_map(self):
        return self._metallic_map
    
    @metallic_map.setter
    @param_setter
    def metallic_map(self, metallic_map:(sampler2D,str,np.ndarray)):
        if isinstance(metallic_map, sampler2D) or metallic_map is None:
            self._metallic_map = metallic_map
        elif isinstance(metallic_map, (str,np.ndarray)):
            if self._metallic_map is None:
                self._metallic_map = sampler2D(metallic_map)
            else:
                self._metallic_map.image = metallic_map

    @property
    def roughness_map(self):
        return self._roughness_map
    
    @roughness_map.setter
    @param_setter
    def roughness_map(self, roughness_map:(sampler2D,str,np.ndarray)):
        if isinstance(roughness_map, sampler2D) or roughness_map is None:
            self._roughness_map = roughness_map
        elif isinstance(roughness_map, (str,np.ndarray)):
            if self._roughness_map is None:
                self._roughness_map = sampler2D(roughness_map)
            else:
                self._roughness_map.image = roughness_map

    @checktype
    def set_as(self, type:Type):
        if type == Material.Type.Emerald:
            self.ambient = glm.vec3(0.0215, 0.1745, 0.0215)
            self.diffuse = glm.vec3(0.07568, 0.61424, 0.07568)
            self.specular = glm.vec3(0.633, 0.727811, 0.633)
            self.shininess = 0.6*128
        elif type == Material.Type.Jade:
            self.ambient = glm.vec3(0.135, 0.2225, 0.1575)
            self.diffuse = glm.vec3(0.54, 0.89, 0.63)
            self.specular = glm.vec3(0.316228, 0.316228, 0.316228)
            self.shininess = 0.1*128
        elif type == Material.Type.Obsidian:
            self.ambient = glm.vec3(0.05375, 0.05, 0.06625)
            self.diffuse = glm.vec3(0.18275, 0.17, 0.22525)
            self.specular = glm.vec3(0.332741, 0.328634, 0.346435)
            self.shininess = 0.3*128
        elif type == Material.Type.Pearl:
            self.ambient = glm.vec3(0.25, 0.20725, 0.20725)
            self.diffuse = glm.vec3(1, 0.829, 0.829)
            self.specular = glm.vec3(0.296648, 0.296648, 0.296648)
            self.shininess = 0.088*128
        elif type == Material.Type.Ruby:
            self.ambient = glm.vec3(0.1745, 0.01175, 0.01175)
            self.diffuse = glm.vec3(0.61424, 0.04136, 0.04136)
            self.specular = glm.vec3(0.727811, 0.626959, 0.626959)
            self.shininess = 0.6*128
        elif type == Material.Type.Turquoise:
            self.ambient = glm.vec3(0.1, 0.18725, 0.1745)
            self.diffuse = glm.vec3(0.396, 0.74151, 0.69102)
            self.specular = glm.vec3(0.297254, 0.30829, 0.306678)
            self.shininess = 0.1*128
        elif type == Material.Type.Brass:
            self.ambient = glm.vec3(0.329412, 0.223529, 0.027451)
            self.diffuse = glm.vec3(0.780392, 0.568627, 0.113725)
            self.specular = glm.vec3(0.992157, 0.941176, 0.807843)
            self.shininess = 0.21794872*128
        elif type == Material.Type.Bronze:
            self.ambient = glm.vec3(0.2125, 0.1275, 0.054)
            self.diffuse = glm.vec3(0.714, 0.4284, 0.18144)
            self.specular = glm.vec3(0.393548, 0.271906, 0.166721)
            self.shininess = 0.2*128
        elif type == Material.Type.Chrome:
            self.ambient = glm.vec3(0.25, 0.25, 0.25)
            self.diffuse = glm.vec3(0.4, 0.4, 0.4)
            self.specular = glm.vec3(0.774597, 0.774597, 0.774597)
            self.shininess = 0.6*128
        elif type == Material.Type.Copper:
            self.ambient = glm.vec3(0.19125, 0.0735, 0.0225)
            self.diffuse = glm.vec3(0.7038, 0.27048, 0.0828)
            self.specular = glm.vec3(0.256777, 0.137622, 0.086014)
            self.shininess = 0.1*128
        elif type == Material.Type.Gold:
            self.ambient = glm.vec3(0.24725, 0.1995, 0.0745)
            self.diffuse = glm.vec3(0.75164, 0.60648, 0.22648)
            self.specular = glm.vec3(0.628281, 0.555802, 0.366065)
            self.shininess = 0.4*128
        elif type == Material.Type.Silver:
            self.ambient = glm.vec3(0.19225, 0.19225, 0.19225)
            self.diffuse = glm.vec3(0.50754, 0.50754, 0.50754)
            self.specular = glm.vec3(0.508273, 0.508273, 0.508273)
            self.shininess = 0.4*128
        elif type == Material.Type.BlackPlastic:
            self.ambient = glm.vec3(0.0, 0.0, 0.0)
            self.diffuse = glm.vec3(0.01, 0.01, 0.01)
            self.specular = glm.vec3(0.50, 0.50, 0.50)
            self.shininess = 0.25*128
        elif type == Material.Type.CyanPlastic:
            self.ambient = glm.vec3(0.0, 0.1, 0.06)
            self.diffuse = glm.vec3(0.0, 0.50980392, 0.50980392)
            self.specular = glm.vec3(0.50196078, 0.50196078, 0.50196078)
            self.shininess = 0.25*128
        elif type == Material.Type.GreenPlastic:
            self.ambient = glm.vec3(0.0, 0.0, 0.0)
            self.diffuse = glm.vec3(0.1, 0.35, 0.1)
            self.specular = glm.vec3(0.45, 0.55, 0.45)
            self.shininess = 0.25*128
        elif type == Material.Type.RedPlastic:
            self.ambient = glm.vec3(0.0, 0.0, 0.0)
            self.diffuse = glm.vec3(0.5, 0.0, 0.0)
            self.specular = glm.vec3(0.7, 0.6, 0.6)
            self.shininess = 0.25*128
        elif type == Material.Type.WhitePlastic:
            self.ambient = glm.vec3(0.0, 0.0, 0.0)
            self.diffuse = glm.vec3(0.55, 0.55, 0.55)
            self.specular = glm.vec3(0.70, 0.70, 0.70)
            self.shininess = 0.25*128
        elif type == Material.Type.YellowPlastic:
            self.ambient = glm.vec3(0.0, 0.0, 0.0)
            self.diffuse = glm.vec3(0.5, 0.5, 0.0)
            self.specular = glm.vec3(0.60, 0.60, 0.50)
            self.shininess = 0.25*128
        elif type == Material.Type.BlackRubber:
            self.ambient = glm.vec3(0.02, 0.02, 0.02)
            self.diffuse = glm.vec3(0.01, 0.01, 0.01)
            self.specular = glm.vec3(0.4, 0.4, 0.4)
            self.shininess = 0.078125*128
        elif type == Material.Type.CyanRubber:
            self.ambient = glm.vec3(0.0, 0.05, 0.05)
            self.diffuse = glm.vec3(0.4, 0.5, 0.5)
            self.specular = glm.vec3(0.04, 0.7, 0.7)
            self.shininess = 0.078125*128
        elif type == Material.Type.GreenRubber:
            self.ambient = glm.vec3(0.0, 0.05, 0.0)
            self.diffuse = glm.vec3(0.4, 0.5, 0.4)
            self.specular = glm.vec3(0.04, 0.7, 0.04)
            self.shininess = 0.078125*128
        elif type == Material.Type.RedRubber:
            self.ambient = glm.vec3(0.05, 0.0, 0.0)
            self.diffuse = glm.vec3(0.5, 0.4, 0.4)
            self.specular = glm.vec3(0.7, 0.04, 0.04)
            self.shininess = 0.078125*128
        elif type == Material.Type.WhiteRubber:
            self.ambient = glm.vec3(0.05, 0.05, 0.05)
            self.diffuse = glm.vec3(0.5, 0.5, 0.5)
            self.specular = glm.vec3(0.7, 0.7, 0.7)
            self.shininess = 0.078125*128
        elif type == Material.Type.YellowRubber:
            self.ambient = glm.vec3(0.05, 0.05, 0.0)
            self.diffuse = glm.vec3(0.5, 0.5, 0.4)
            self.specular = glm.vec3(0.7, 0.7, 0.04)
            self.shininess = 0.078125*128
    
    def _update_all_depth_maps(self):
        for mesh in self._parent_meshes:
            for scene in mesh.scenes:
                scene._should_update_depth_maps = True

    def _update_all_env_maps(self):
        for mesh in self._parent_meshes:
            for scene in mesh.scenes:
                scene._should_update_env_maps = True

    def _test_transparent(self):
        self._has_transparent = False
        self._has_opaque = False
        if self.opacity_map is not None:
            image = self.opacity_map.image
            if image is None:
                self._has_transparent = True
                self._has_opaque = True
            else:
                if "int" in str(image.dtype):
                    self._has_transparent = np.any(image < 255)
                    self._has_opaque = np.any(image >= 255)
                else:
                    self._has_transparent = np.any(image < 1-1E-6)
                    self._has_opaque = np.any(image >= 1-1E-6)
        else:
            self._has_transparent = (self.opacity < 1-1E-6)
            self._has_opaque = (self.opacity >= 1-1E-6)

        if self.diffuse_map is not None:
            image = self.diffuse_map.image
            if image is None:
                self._has_transparent = True
                self._has_opaque = True
            elif len(image.shape) > 2 and image.shape[2] > 3:
                if "int" in str(image.dtype):
                    self._has_transparent = (np.any(image[:,:,3] < 255) or self._has_transparent)
                    self._has_opaque = (np.any(image[:,:,3] >= 255) and self._has_opaque)
                else:
                    self._has_transparent = (np.any(image[:,:,3] < 1-1E-6) or self._has_transparent)
                    self._has_opaque = (np.any(image[:,:,3] >= 1-1E-6) and self._has_opaque)

        for mesh in self._parent_meshes:
            mesh._test_transparent()


from .SceneNode import SceneNode
from .Manipulators.SceneRoamManipulator import SceneRoamManipulator
from .Renderers.ForwardRenderer import ForwardRenderer
from .VideoRecorder import VideoRecorder

from glass.utils import checktype

import glm
import math
from enum import Enum
import numpy as np
import sys

Screen = {}
def import_Screen(gui_system:str)->None:
    if gui_system not in Screen:
        cmd = f'from .Screens.{gui_system}Screen import {gui_system}Screen'
        exec(cmd)

        Screen[gui_system] = eval(f"{gui_system}Screen")

class Camera(SceneNode):

    class ProjectionMode(Enum):
        Perspective = 0
        Orthographic = 1

    class Lens:

        def __init__(self):
            self.focus:float = 0.09
            self.aperture:float = 0.05
            self.auto_focus:bool = True
            self.focus_tex_coord:glm.vec2 = glm.vec2(0.5, 0.5)
            self.focus_change_time:float = 2

            self.explosure:float = 1
            self.auto_explosure:bool = True
            self.local_explosure:bool = False
            self.explosure_adapt_time:float = 2

        @property
        def clear_distance(self):
            return 1/(1/self.focus - 1/self.near)
        
        @clear_distance.setter
        def clear_distance(self, distance:float):
            self.focus = 1/(1/self.near + 1/distance)

    @checktype
    def __init__(self, gui_system="", projection_mode:ProjectionMode=ProjectionMode.Perspective, name:str=""):
        SceneNode.__init__(self, name, unique_path=True)
        self.__projection_mode:Camera.ProjectionMode = projection_mode
        self.__gui_system:str = ""

        self.__fov_deg:float = 45
        self.__fov_rad:float = math.pi/4
        half_fov = self.__fov_rad/2
        self.__tan_half_fov:float = math.tan(half_fov)
        self.__sin_half_fov:float = math.sin(half_fov)
        self.__far:float = 100
        self.__near:float = 0.1
        self.__clip:float = self.__far - self.__near
        self.__height:float = 40*2*self.__near*self.__tan_half_fov
        self.__CSM_levels:int = 5
        self.__lens:Camera.Lens = Camera.Lens()

        self._set_screen(gui_system)

    def _set_screen(self, gui_system)->None:
        if not isinstance(gui_system, str):
            gui_system = gui_system.__name__

        if not gui_system:
            for module_name in sys.modules:
                if module_name.startswith("PyQt6"):
                    gui_system = "PyQt6"
                    break

                if module_name.startswith("PyQt5"):
                    gui_system = "PyQt5"
                    break

                if module_name.startswith("PySide6"):
                    gui_system = "PySide6"
                    break

                if module_name.startswith("PySide2"):
                    gui_system = "PySide2"
                    break

        if gui_system:
            self.__gui_system = gui_system.lower().replace("pyside", "PySide").replace("pyqt", "PyQt")
            import_Screen(self.__gui_system)

        if not self.__gui_system:
            try:
                import_Screen("PyQt6")
                self.__gui_system = "PyQt6"
            except:
                self.__gui_system = ""

        if not self.__gui_system:
            try:
                import_Screen("PyQt5")
                self.__gui_system = "PyQt5"
            except:
                self.__gui_system = ""

        if not self.__gui_system:
            try:
                import_Screen("PySide6")
                self.__gui_system = "PySide6"
            except:
                self.__gui_system = ""

        if not self.__gui_system:
            try:
                import_Screen("PySide2")
                self.__gui_system = "PySide2"
            except:
                self.__gui_system = ""

        if not self.__gui_system:
            raise RuntimeError("all GUI systems failed to init")

        self.__screen = Screen[self.__gui_system](self)
        self.__screen.manipulator = SceneRoamManipulator()
        self.__screen.renderer = ForwardRenderer()

    @property
    def lens(self)->Lens:
        return self.__lens

    @property
    def gui_system(self)->str:
        return self.__gui_system
    
    @lens.setter
    def lens(self, lens:Lens)->None:
        self.__lens = lens

    @property
    def projection_mode(self)->ProjectionMode:
        return self.__projection_mode
    
    @projection_mode.setter
    def projection_mode(self, projection_mode:ProjectionMode)->None:
        self.__projection_mode = projection_mode

    @property
    def tan_half_fov(self)->float:
        return self.__tan_half_fov
    
    @property
    def sin_half_fov(self)->float:
        return self.__sin_half_fov

    @property
    def aspect(self)->float:
        return self.__screen.width() / self.__screen.height()
    
    @property
    def fov_x(self)->float:
        return 2*math.atan(self.aspect*self.__tan_half_fov)/math.pi*180
    
    @property
    def fov_y(self)->float:
        return self.__fov_deg
    
    @property
    def fov(self)->float:
        return self.__fov_deg
    
    @fov.setter
    def fov(self, fov_deg:float)->None:
        self.__fov_deg = fov_deg
        self.__fov_rad = fov_deg/180*math.pi
        half_fov = self.__fov_rad/2
        self.__tan_half_fov = math.tan(half_fov)
        self.__sin_half_fov = math.sin(half_fov)
        self.__height = 40*2*self.__near*self.__tan_half_fov

    @property
    def near(self)->float:
        return self.__near
    
    @near.setter
    def near(self, near:float)->None:
        self.__near = near
        self.__clip = self.__far - self.__near
        self.__height = 40*2*self.__near*self.__tan_half_fov
    
    @property
    def far(self)->float:
        return self.__far
    
    @far.setter
    def far(self, far:float)->None:
        self.__far = far
        self.__clip = self.__far - self.__near

    @property
    def clip(self)->float:
        return self.__clip

    @property
    def height(self)->float:
        return self.__height

    @height.setter
    def height(self, height:float)->None:
        self.__height = height

        self.__tan_half_fov = self.__height / (40*2*self.__near)
        half_fov = math.atan(self.__tan_half_fov)
        self.__fov_rad = 2*half_fov
        self.__fov_deg = self.__fov_rad/math.pi*180
        self.__sin_half_fov = math.sin(half_fov)

    @property
    def width(self)->float:
        return self.__height * self.aspect
    
    @width.setter
    def width(self, width:float)->None:
        self.__height = width / self.screen.width() * self.screen.height()

    @property
    def screen(self):
        return self.__screen

    @property
    def CSM_levels(self)->int:
        return self.__CSM_levels
    
    @CSM_levels.setter
    def CSM_levels(self, levels:int):
        self.__CSM_levels = levels

    def take_photo(self, save_path:str=None, viewport:tuple=None)->np.ndarray:
        return self.screen.capture(save_path, viewport)
    
    def record_video(self, save_path:str, viewport:tuple=None, fps:(float,int)=None)->VideoRecorder:
        return self.screen.capture_video(save_path, viewport, fps)

    def project(self, world_coord:glm.vec3)->glm.vec4:
        return self.view_to_NDC(self.world_to_view(world_coord))
    
    def project3(self, world_coord:glm.vec3)->glm.vec3:
        NDC = self.project(world_coord)
        return NDC.xyz / NDC.w

    def world_to_view(self, world_coord:glm.vec3)->glm.vec3:
        return glm.inverse(self.abs_orientation) * (world_coord - self.abs_position)

    def view_to_world(self, view_coord:glm.vec3)->glm.vec3:
        return self.abs_orientation * view_coord + self.abs_position

    def world_dir_to_view(self, world_dir:glm.vec3)->glm.vec3:
        return glm.inverse(self.abs_orientation) * world_dir

    def view_dir_to_world(self, view_dir:glm.vec3)->glm.vec3:
        return self.abs_orientation * view_dir
    
    def view_to_NDC(self, view_coord:glm.vec3)->glm.vec4:
        NDC_coord = glm.vec4()
        if self.projection_mode == Camera.ProjectionMode.Perspective:
            NDC_coord.x = view_coord.x / (self.aspect * self.tan_half_fov)
            NDC_coord.y = view_coord.z / self.tan_half_fov
            NDC_coord.z = 2*self.far*(view_coord.y-self.near)/self.clip-view_coord.y
            NDC_coord.w = view_coord.y
        else:
            NDC_coord.x = 2*view_coord.x / self.width
            NDC_coord.y = 2*view_coord.z / self.height
            NDC_coord.z = 2*(view_coord.y-self.near)/self.clip-1
            NDC_coord.w = 1
        
        return NDC_coord
    
    def screen_to_view_dir(self, screen_pos:glm.vec2)->glm.vec3:
        xNDC = 2*screen_pos.x / self.screen.width() - 1
        yNDC = 1 - 2*screen_pos.y / self.screen.height()

        view_dir = glm.vec3()
        view_dir.x = xNDC * self.aspect * self.tan_half_fov
        view_dir.y = 1
        view_dir.z = yNDC * self.tan_half_fov
        return glm.normalize(view_dir)
    
    def screen_to_world_dir(self, screen_pos:glm.vec2)->glm.vec3:
        return self.view_dir_to_world(self.screen_to_view_dir(screen_pos))
    


from abc import ABC, abstractmethod

from glass.utils import di

class Renderer(ABC):

    def __init__(self):
        self._camera_id = id(None)

    @property
    def camera(self): 
        return di(self._camera_id)

    @property
    def scene(self):        
        return self.camera.scene
        
    def startup(self):
        pass

    @abstractmethod
    def render(self)->bool:
        pass



from .BO import BO

from OpenGL import GL

class EBO(BO):

	_basic_info = \
	{
		"gen_func": GL.glGenBuffers,
		"bind_func": GL.glBindBuffer,
		"del_func": GL.glDeleteBuffers,
		"target_type": GL.GL_ELEMENT_ARRAY_BUFFER,
		"binding_type": GL.GL_ELEMENT_ARRAY_BUFFER_BINDING,
		"need_number": True,
	}

	def __init__(self):
		BO.__init__(self)


import sys
import os

from glass.download import pip_install, is_China_user, download

try:
    import PyQt6
except:
    pip_install("PyQt6")

from PyQt6.QtWidgets import QApplication, QDialog, QHBoxLayout, QVBoxLayout, QWidget
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QIcon
from qt_material import apply_stylesheet

from .QtComponents import SliderInput

from ..Camera import Camera
from ..Model import Model
from ..BasicScene import ModelView
from ..Geometries.CoordSys import CoordSys

class TransformControlPanel(QWidget):

    def __init__(self, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)
        hlayout = QHBoxLayout()

        vlayout = QVBoxLayout()
        self.position_x_line = SliderInput("position.x", (-5, 5), 0, "m", self)
        vlayout.addWidget(self.position_x_line)

        self.position_y_line = SliderInput("position.y", (-5, 5), 0, "m", self)
        vlayout.addWidget(self.position_y_line)

        self.position_z_line = SliderInput("position.z", (-5, 5), 0, "m", self)
        vlayout.addWidget(self.position_z_line)
        vlayout.addStretch()

        hlayout.addLayout(vlayout)

        vlayout = QVBoxLayout()
        self.yaw_line = SliderInput("偏航角 (yaw)", (-180, 180), 0, "°", self)
        vlayout.addWidget(self.yaw_line)

        self.pitch_line = SliderInput("俯仰角 (pitch)", (-90, 90), 0, "°", self)
        vlayout.addWidget(self.pitch_line)

        self.roll_line = SliderInput("翻滚角 (roll)", (-180, 180), 0, "°", self)
        vlayout.addWidget(self.roll_line)
        vlayout.addStretch()

        hlayout.addLayout(vlayout)

        vlayout = QVBoxLayout()
        self.scale_x_line = SliderInput("scale.x", (-2, 2), 1, "", self)
        vlayout.addWidget(self.scale_x_line)

        self.scale_y_line = SliderInput("scale.y", (-2, 2), 1, "", self)
        vlayout.addWidget(self.scale_y_line)

        self.scale_z_line = SliderInput("scale.z", (-2, 2), 1, "", self)
        vlayout.addWidget(self.scale_z_line)
        vlayout.addStretch()

        hlayout.addLayout(vlayout)

        self.setLayout(hlayout)

class MainWindow(QDialog):

    def __init__(self, parent:QWidget=None):
        QDialog.__init__(self, parent=parent)
        self.setWindowFlags(Qt.WindowType.WindowMinMaxButtonsHint | Qt.WindowType.WindowCloseButtonHint)
        self.setWindowTitle('glass_engine 空间变换概念演示')

        self_folder = os.path.dirname(os.path.abspath(__file__))
        self.setWindowIcon(QIcon(self_folder + "/../images/glass_engine_logo64.png"))

        vlayout = QVBoxLayout()

        self.scene, self.camera, dir_light = ModelView()
        self.camera.screen.manipulator.elevation = 30
        self.camera.screen.manipulator.distance = 5
        self.camera.projection_mode = Camera.ProjectionMode.Orthographic

        dir_light.generate_shadows = False

        self_path = os.path.dirname(os.path.abspath(__file__))
        self.model = Model(self_path + "/assets/models/jet/11805_airplane_v2_L2.obj")
        self.model["root"].scale = 1/500
        self.model["root"].position.z -= 0.3
        model_coord_sys = CoordSys(1.6)
        self.model.add_child(model_coord_sys)

        coord_sys = CoordSys(1.6, alpha=0.3)

        self.scene.add(self.model)
        self.scene.add(coord_sys)

        self.scene.skydome = "https://dl.polyhaven.org/file/ph-assets/HDRIs/extra/Tonemapped%20JPG/industrial_sunset_puresky.jpg"
        
        vlayout.addWidget(self.camera.screen)

        self.control_panel = TransformControlPanel(self)
        vlayout.addWidget(self.control_panel)

        self.setLayout(vlayout)
        self.resize(1000, 600)

        self.control_panel.position_x_line.value_changed.connect(self.position_x_changed)
        self.control_panel.position_y_line.value_changed.connect(self.position_y_changed)
        self.control_panel.position_z_line.value_changed.connect(self.position_z_changed)
        self.control_panel.yaw_line.value_changed.connect(self.yaw_changed)
        self.control_panel.pitch_line.value_changed.connect(self.pitch_changed)
        self.control_panel.roll_line.value_changed.connect(self.roll_changed)
        self.control_panel.scale_x_line.value_changed.connect(self.scale_x_changed)
        self.control_panel.scale_y_line.value_changed.connect(self.scale_y_changed)
        self.control_panel.scale_z_line.value_changed.connect(self.scale_z_changed)
        self.camera.screen.setFocus()
        
    def position_x_changed(self, value):
        self.model.position.x = value
        self.camera.screen.update()

    def position_y_changed(self, value):
        self.model.position.y = value
        self.camera.screen.update()

    def position_z_changed(self, value):
        self.model.position.z = value
        self.camera.screen.update()

    def yaw_changed(self, value):
        self.model.yaw = value
        self.camera.screen.update()

    def pitch_changed(self, value):
        self.model.pitch = value
        self.camera.screen.update()

    def roll_changed(self, value):
        self.model.roll = value
        self.camera.screen.update()

    def scale_x_changed(self, value):
        self.model.scale.x = value
        self.camera.screen.update()

    def scale_y_changed(self, value):
        self.model.scale.y = value
        self.camera.screen.update()

    def scale_z_changed(self, value):
        self.model.scale.z = value
        self.camera.screen.update()

def download_files():
    self_folder = os.path.dirname(os.path.abspath(__file__)).replace("\\", "/")
    model_folder = self_folder + "/assets/models/jet"

    md5_map = \
    {
        "11805_airplane_v2_L2.mtl": "36059ea10e4a35955dc2ee6b1203f9c8",
        "11805_airplane_v2_L2.obj": "85bd6ac8fcc6c717ba5a96a7a065c0ec",
        "airplane_body_diffuse_v1.jpg": "e7b1c3492a69e81959ffc14af0ad8ed7",
        "airplane_wings_diffuse_v1.jpg": "c4ef09fd0825d5d97a1ba105775e253a",
    }

    target_file = model_folder + "/11805_airplane_v2_L2.mtl"
    gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/Demos/assets/models/jet/11805_airplane_v2_L2.mtl"
    github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/Demos/assets/models/jet/11805_airplane_v2_L2.mtl"

    if is_China_user():
        download(gitee_url, target_file, md5_map["11805_airplane_v2_L2.mtl"])
    else:
        download(github_url, target_file, md5_map["11805_airplane_v2_L2.mtl"])

    target_file = model_folder + "/11805_airplane_v2_L2.obj"
    gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/Demos/assets/models/jet/11805_airplane_v2_L2.obj"
    github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/Demos/assets/models/jet/11805_airplane_v2_L2.obj"

    if is_China_user():
        download(gitee_url, target_file, md5_map["11805_airplane_v2_L2.obj"])
    else:
        download(github_url, target_file, md5_map["11805_airplane_v2_L2.obj"])

    target_file = model_folder + "/airplane_body_diffuse_v1.jpg"
    gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/Demos/assets/models/jet/airplane_body_diffuse_v1.jpg"
    github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/Demos/assets/models/jet/airplane_body_diffuse_v1.jpg"

    if is_China_user():
        download(gitee_url, target_file, md5_map["airplane_body_diffuse_v1.jpg"])
    else:
        download(github_url, target_file, md5_map["airplane_body_diffuse_v1.jpg"])

    target_file = model_folder + "/airplane_wings_diffuse_v1.jpg"
    gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/glass_engine/Demos/assets/models/jet/airplane_wings_diffuse_v1.jpg"
    github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/glass_engine/Demos/assets/models/jet/airplane_wings_diffuse_v1.jpg"

    if is_China_user():
        download(gitee_url, target_file, md5_map["airplane_wings_diffuse_v1.jpg"])
    else:
        download(github_url, target_file, md5_map["airplane_wings_diffuse_v1.jpg"])

def demo_transform():
    download_files()
    app = QApplication(sys.argv)
    apply_stylesheet(app, theme='dark_amber.xml')

    main_window = MainWindow()
    main_window.show()

    app.exec()


from .ColorInput import ColorChooser, ColorInput
from .ImageChooser import ImageChooser
from .MapInput import MapInput
from .SliderInput import SliderInput
from .ComboboxInput import ComboboxInput
from .ColorImageInput import ColorImageInput, ColorImageChooser
from .ValueImageInput import ValueImageInput
from .IntInput import IntInput


from .GLInfo import GLInfo
from .GLObject import GLObject
from .TextureUnits import TextureUnits
from .GLConfig import GLConfig
from .utils import checktype

import random
from functools import wraps

class FBOAttachment(GLObject):

    def __init__(self):
        GLObject.__init__(self)

        self._fbo = None
        self._fbo_attach_point = None
        self._fbo_image_changed = False
        self._fbo_image_generated_mipmap = True
        self._can_resize = True
        self._dynamic = True

    @property
    def dynamic(self):
        return self._dynamic

    @staticmethod
    def param_setter(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            value = args[1]

            equal = False
            try:
                lvalue = getattr(self, func.__name__)
                if type(lvalue) != type(value):
                    equal = False
                else:
                    equal = bool(getattr(self, func.__name__) == value)
            except:
                equal = False

            if equal:
                return

            if not self.dynamic:
                raise RuntimeError(f"none dynamic {self.__class__.__name__} cannot change any parameters")

            safe_func = checktype(func)
            return_value = safe_func(*args, **kwargs)

            return return_value

        return wrapper

    def malloc(self, width:int, height:int, samples:int=None, layers:int=None, internal_format:GLInfo.internal_formats=None):
        pass

    def resize(self, width:int, height:int, samples:int=None, layers:int=None):
        if not self._can_resize:
            return

        self._can_resize = False

        self.malloc(width, height, samples, layers, internal_format=None)
        if self._fbo is not None:
            self._fbo.resize(width, height, samples, layers)

        self._can_resize = True

    def clear_buffer(self):
        self.bind(force_update_image=True)

    @property
    def fbo(self):
        return self._fbo

    def bind(self, update_fbo:bool=False, force_update_image:bool=False):
        cls_name = self.__class__.__name__
        target_type = self.__class__._basic_info["target_type"]
        if "sampler" in cls_name:
            texture_unit = None
            if self._id == 0:
                texture_unit = TextureUnits.available_unit
            else:
                texture_unit = TextureUnits.unit_of_texture((target_type, self._id))
                if texture_unit is None:
                    texture_unit = TextureUnits.available_unit

            if texture_unit is None:
                texture_unit = random.randint(0, GLConfig.max_texture_units-1)

            GLConfig.active_texture_unit = texture_unit
            
        GLObject.bind(self)
        if update_fbo:
            self._fbo_image_changed = True
            self._fbo_image_generated_mipmap = False
        
        if "sampler" in cls_name:
            TextureUnits[texture_unit] = (target_type, self._id)

    def unbind(self):
        success = GLObject.unbind(self)
        cls_name = self.__class__.__name__
        if success and "sampler" in cls_name:
            target_type = self.__class__._basic_info["target_type"]
            assert TextureUnits.current_texture == (target_type, self._id)
            TextureUnits.current_texture = (target_type, 0)

        return success
    
    def __del__(self):
        self.unbind()
        GLObject.__del__(self)


from .Cylinder import Cylinder
from .Cone import Cone
from ..SceneNode import SceneNode

import glm

class CoordSys(SceneNode):

    def __init__(self, x_length:float=1, y_length:float=None, z_length:float=None, alpha:float=1, name:str=""):
        SceneNode.__init__(self, name=name)

        if y_length is None:
            y_length = x_length

        if z_length is None:
            z_length = x_length

        x_axis = CoordSys.create_axis(x_length, glm.vec4(1, 0, 0, alpha))
        x_axis.roll = 90
        self.add_child(x_axis)

        y_axis = CoordSys.create_axis(y_length, glm.vec4(0, 1, 0, alpha))
        y_axis.pitch = -90
        self.add_child(y_axis)

        z_axis = CoordSys.create_axis(z_length, glm.vec4(0, 0, 1, alpha))
        self.add_child(z_axis)

    @staticmethod
    def create_axis(length:float, color:glm.vec4):
        axis = Cylinder(0.015, length, color=color)
        arrow = Cone(0.04, 0.15, color=color)
        arrow.position.z = length
        axis.add_child(arrow)

        return axis